 /* aaaaaaaaaaaaa   aaaaa     aaaa	aaaaaaaaaaaaaaaaa
H3sm 1.4.m4
see, $segments...

H3m 3-stack programming interface (virtual machine design) in GNU C, m4,
	and GNU "as" assembler directives for x86.

199804 <--> 200003   Rick Hohensee
humbubba@smart.net	http://www.smart.net/~humbubba

199812	 Michael Somos "http://grail.cba.csuohio.edu/~somos/"
	non-endianness, routine delimiter comment lines, etc.

See also Ertl/Paysan GForth and GNU .info on gcc "labels as values",
the unix section 2 manpages, Linux man 2 intro in particular, and the
GNU .info on gas.

The sourcecode of the "Arith" and "calc" arbitrary precision arithmetic
packages for unices were consulted for this implementation of pyte
arithmetic, although pytes are quite different from numbers in those
systems.

Phil Koopman's book "Stack Computers; The New Wave" bears mention. It is
currently on the web at "http://www.cs.cmu.edu/~koopman/stack_computers/".

Also worth mention are the generally meritocratic and cooperative tones of
comp.lang.forth, MdFIG and linux-kernel@vger.rutgers.edu. Greg Patterson
was helpful at a particularly frustrating moment with C pointers. Two
design features of H3m were suggested by JET Thomas; the single-whole-pass
dictionary search method used by abbreviations, and that pyte sizes go up
to 256 bytes instead of 128. Explanations by Elizabeth Rather and others
in c.l.f led to me eliminating H3m words that access the parameter stack
pointer directly, since a pure stack that has no concept of a "stack
pointer" is sufficient for a stack machine, and probably preferable.  I
had a copy of John Brien's "Build Your Own Forth" webpage in my
/usr/src/H3m . (original URL is
"http://www.users.zetnet.co.uk/aborigine/byof.htm")

The printer Charles Lasitter of Durham, N.C. gave me has done yeoman
service as an essential debugging tool. As have the several PS/2's,
the Pentium, and the cash. I believe the technical term for the likes
of Chuck Lasitter is "angel".

H3m has also recieved direct help from Richard B. Johnson, Brian Raiter,
and direct help or inspiration from various anonymous Internet Relay Chat
personages. The nom_de_IRC's _Anarchy_, tm2, shaver and * come to mind.

As pertains copyrights and release thereof, H3m is not a derived work of
any prior work.

This version of H3m is hereby released into the public domain. I am of
the understanding that "public domain" does not release users of H3m from
the obligation of properly associating my name with the work. Subsequent
work I do on H3m may be released under more restrictive terms.

DESIGN

Virtual Machine Subroutine Threaded H3m.
	All "executable" dictionary cells contain actual addresses of code.
	The address of a sub-word is in the callER, following a "call"
	execution token cell, which is not present in more prevalent
	threading schemes.

Each of the 3 "stacks" has a distinctly different related ops from the other
2.

return
	The return stack will contain only xt's. An xt is what NEXT can use,
	an address space int-ply index to an atom, which is machine code.

pointer
	The H3m pointer stack is cell-sized, but contains only actual host
	address-space pointers usually. Such an entity is called a "pointer"
	from the H3m-user point of view. The pointer stack may be a bridge
	between the H3m VM and the host, but only with specific fundamentals
	for host services, such as system calls. Internally, the pointer stack
	index is usually left pointing to an occupied cell.
data
	The data stack operates on "pytes", groups of 1 <--> 256
	bytes. Boolean flags are the low byte of a pyte. False is zero.
	Non-zero is true. The H3m   true   word asserts 255 in a flagpyte.
	The stack index itself is byte-ply, but pyte operations are at Size.
	The data stack pointer points at the top occupied cell. This is
	important to keep in mind when changing Size.

	Size "Register"

	The current effective size of data stack operands is the Size state
	variable. There are user-visible accessors of this Size "register".
	Operations on pytes are in terms of Size, except where a pyte is
	treated as a flag or character.

So the 3 stacks are the data typing of H3m, in a sense. Typing is enforced
by the various operands. The data stack is where a datum can be treated
arbitrarily.

IMPLEMENTATION

Early versions may depend on 4 byte address cell.  For big-endian CPU see
"CONFIGURATION SWITCH" below.  Initial implementation. REQUIRES *GNU* C.

This H3m is for x86. The CPU-specific code is not much, but it is spread
around, so I'm not going to attempt to isolate it with the cpp #include
mechanism. If someone ports H3m to another platform I'll include the
whole thing in the package if they want, or a patch. The approximate
minimum machine-specific stuff for a unix would be the the exit, write and
read syscalls, which are the int 0x80 instruction on x86 Linux. This H3m
needs those syscalls for several primitives te be usable at all, and has
the calls coded in situ where needed. After those you'll want argv/env
passing, which requires machine opcodes, and other syscalls. A
per-keystroke input mechanism probably wants ioctl to set the tty to raw
mode, for example. I think.

H3m NEXT uses gcc labels_as_values and computed goto. The Linux syscalls
use gcc C expressions in asm statements. The idiom of this code overall is
what I call <C, less-than C. ip is a C "pointer to int", as needed for
gcc goto **. It's the only pointer declaration in H3m.

As of this writing H3m assumes a 32 bit address cell, but I suspect that
can be macro'ed out for a configurable cell size with little bloodshed.
Some of that has been done already.

C slash-asterisk comments beginning in leftmost character column are
appended to the glossary file by the Glossary script.

OK, let's get coding...	  */
	 /* bytes in an address cell */
#define CELLSIZE     sizeof(int)
#define STACKSIZE	0x2000

#define SUSPECT

 /* declare big empty stuff above main(). This keeps it out of the binary
		object file. (in .bss). This should be possible for
		block and buffer also, with some init code.	 */

typedef unsigned char byte;

int cells;

			/* allocate the front of .bss for heap. It's just 
			there, we don't use it's name. I tried to get some
			big buffers out of the dictionary, and failed, so
			the exe is several k bigger on disk than it might 
			otherwise be, but things are where thier names are 
			physically as per what you might expect.
			*/

asm (" .lcomm HEAP, 0x20000 ");

					/* data byte stack, byte access array */
byte		ds[STACKSIZE];
					/* byte-only data stack, everything gets
					converted to bytes internally in this
					implementation.	 */
int		rs[256];		/* return stack itself, ints = xt's */
int		ps[256];		/* the pointer stack itself, an array of
					 *****ints*******	   */

	/* these declarations are above main() to please arg_env.o
and gcc -nostdlib  */

int argc, i;
char ** argv;
int argspan ;

 /* CONFIGURATION SWITCH  CONFIGURATION SWITCH
endianism stuff, change $1 $2 $3 $4 to $4 $3 $2 $1 for big-endian,
(after	Michael Somos). In re 64-bit etc. I'm open to advice.  */



 /* cpp/C macros for within atoms */

#define DROP	dsi = dsi - Size ; if (dsi < 0) { dsi=0; }
#define BUMP	dsi = dsi + Size ;
#define TRUE	dsi = dsi + Size ; ds[dsi] = 255 ;
#define FALSE	dsi = dsi + Size ; ds[dsi] = 0 ;

 /* comment-anywhere macros */

			/* overwrite TOS with cell above TOS */
#define MASH				\
j = Size;				\
while ( j-- )				\
{ ds[dsi + j] = ds[dsi + j + Size] ; }	\

 /*  ######  pyte arithmatic declarations  #######    */
typedef unsigned short	dual;		/* 16 bits. 2 bytes.  */

byte bytes[512], modulo[256];
dual accum[256];
dual head;

 /* digits up to base 16. I prefer lowercase. Also, e.g. FF is then a WORD
only. Handy. */
char  digitOut[] = {	'0','1','2','3','4','5','6','7','8','9',
			'a','b','c','d','e','f'	 };

char numberString[512];		/* to save syscalling write on each digit */

 /* convert.[bi] bytes-to-int zone */
union  intbyte	{
	int    i;
	byte   b[CELLSIZE];
		} convert;

 /* MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN
H3m is all in one C "function".	  */

char main(){	   /*

With gcc -nostdlib we get a stdIO, argv, envp...
If you want to exit cleanly you call the exit syscall, AKA quit.
There's the C-time init stuff here, and then a branch over all the
dictionary code to a NEXT that "increments ip" to H3m-time. */

int  Size = CELLSIZE  ;	 /* instruction pointer, program counter,
				what NEXT jumps to, and post-increments.
				Register declarations (ip in esp, dsi in
				ebp) I've tried weren't worthwhile.	 */

 int * ip  ;	/* instruction pointer, see NEXT */

int		dsi;
int		rsi;			/* return stack index */
int		psi;			/* pointer stack index */

int		scratch;

int ds0 = (int)&ds;	/* int faux pointer to ds array. GNU cc only? */
int rs0 = (int)&rs;

byte signA, signB ;		  /* signs of args and result */

int   gross;
byte  net;

 /*		NEXT NEXT NEXT NEXT NEXT NEXT NEXT
 In a real CPU the instruction pointer is incremented over code
automatically by every instruction. In a software Forthoid, the NEXT code
snippet at the end of each primitive does this. This is also called the
"inner interpreter", or address interpreter. H3m NEXT uses GNU cc
computed goto. ip is post-incremented.	When in a word, ip is pointing at
next word in calling thread. It occurs to me that inlining could be easy.
Just sling code together and lose NEXT. LAAETTR.  */

#define NEXT   goto **ip++;

int i, j, k, K, e ;
byte bite;

ps[psi] = 0xc001c001;			/* base of ps marker, psi 0 */

 /* set commandline args as H3m parse bounds. H3m token works vis-a-vis
	unix "whitespace".						*/

*(int *)&&epal	  = (int)argv[argc+1] ;
*(int *)&&thisl	   = (int)argv[0];

 /* basic variable initialization		*/
*(int *)&&dpaddy = (int)&&initial_HERE + 1000 ;
*(int *)&&latest = (int)&&initial_LATEST ;

 /* open "stdaux" to tty			*/

__asm__	 ("int $0x80"
	: "=a" (convert.i)
	: "0" (5),"b" ("/dev/tty"),"c" (0),
	  "d" (0)  );

 /* we go from sequential C to NEXT-threaded H3m in 2 moves.		*/
ip = &&zygote ;
 /* ip now contains the address of the HNC of the word "zygote" .	*/
 /* Later the init code does a NEXT like it's a H3m atom and
	it's H3m-time.				*/

goto initial_HERE;

 /* The atoms are the actual runtime machine code of H3m. They are what
the NEXTs call, once H3m itself is running.  The following code has to be
*compiled* here, but where it gets called from and where it jumps to is
what the labels are for, so we skip it for now with a goto.  Everything we
need the real machine for should be in here.  We use gotos ( in NEXT
primarily), but in the manner of _uncooked_ spaghetti, i.e. the gotos of
the NEXTs of each word are laminar, not tangled. There are no cross-word
gotos or fall-throughts. Fall-throughs would be OK, but aren't necessary.
This breaks gcc jump-optimization, e.g. gcc -O1 will remove all the atoms.
H3m is almost un-optimizable anyway though, AFAICT. Apparently gcc produces
code by default that e.g. peephole optimizing can't improve on much.

For some words that need stack items they don't modify, I use
( required ||| consumed --- produced )	 e.g.	  over	  ( a b ||| --- a )

"|null" means "or no stack item". I wish stack comments were concise, but no
guarantees. Running stack comments within words are particularly ad-hoc.

!SIZE  means word probably modifies Size.

AARDVARK!  AARDVARK!  aardvark exists just to be un-'-able due to 0 count
byte, ie. no name, and/or code field, and to be the starting point for the
dictionary linking. Handrolled ( non-macro )  empty neck.
The way to terminate a namespace traversal is when an HNC has a 0 length
byte. :NONAME isn't a problem because that's not a linked thing at all,
is it?
*/

asm	("
	1:
	0:
	.byte 0
	.byte 0
	.byte 0
	.byte 0
	.ascii \"aardvark\"
	.align 4, 0
	.int  0
	.int  0
	");
aardvark:

 /* m4m4m4m4m4m4m4m4m4m4m4 m4 macro to do dictionary things as assembly
directives.  Doing this as macros in cpp doesn't appear to be possible due
to " issues. The semantic overhead of an m4 pass is well paid for by the
strikingly Forthish simplicity of doing what we actually want to do
directly.

There's some running m4 logic to toggle the labels produced in each header
due to some rigidity of gas.  Thanks to Andi Kleen for the general idea.
He further suggests using a counter instead of a mere toggle, which may
come into play with multiple vocabularies, I guess. I'm still open to
simpler methods. I'm not dead sure gas can't do this on it's own, even.
 */




 /* these need to take a C/asm name and a H3m name.
	These could be hand-adjusted for CELL other than 4  */






 /* The m4 pass is first. In most of these macros we compute the length of
the word name at m4-time, which means we need another macro if we need
escapes for .ascii's in asm, which look like name chars to m4. This is the
case for H3m names containing " . */



 /* this is a thread header vith a non-user bit for leaving
interpreter-only  words and such out of "words", i.e. mildly hidden. */



 /* "threading" words are analagous to Forth immediate words */



 /*
real fundamentals/primitives...
possible other primitives; move, fill, endianism converters, charsearch
 aaaaaaaaaaaaaaaaaaaaaaaaaaaa================= */

 /* comments formatted for Glossary ggggggggggggggggggggGGGGGGGGGGGgggg */
/* glossary may include   
asm	("
	1:
	.byte 0
	.byte 0
	.byte 0
	.byte 1
	.ascii \"\"
	.align 4, 0
	.int  0b
	.equ	CFA, .
	");
 words "words" hides.			*/
/* H3m is case-sensitive. Booleans OR, AND, XOR and NOT are uppercase.	*/
/* H3m abbreviation/pronounciation  conventions:			*/
/* */
/*	@	!	->	.	&	'	$		*/
/*	fetch	store	to	print	append	tick	segment		*/
/* */
/* & (append) is analagous to Forth comma.				*/
/* @ is usually a copy, -> is usually a move or pop, source is dropped. */
/* & in stack comments, e.g. &HNC, sometimes means "address of"		*/
/* ||| means "stuff to the left is required but not changed".		*/
/* if* words like if, ifchar etc. produce flagpytes.			*/
/* 		(optional fields, optional order)		*/
/* name		pronounced   stacks effects	 "parsing syntax"   other  */
/* */
/*	ATOMS... */

 /*============================================= */
 /*================= visible dictionary ============================ */


/* address	(P --- ptr ) !rsi  action of named data, Forth VARIABLE */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"address\"
	.align	4, 0
	.int	1b
	.equ	addressCFA, .
	");
address:
		      /*      (P --- ptr )

Push our DFA address. analogue of Forth "dovar", but to pointer stack */

 /* threading issue */
ps[++psi] = (int) ip ;	 /* pointer stack is pre-increment. */
ip = (int *) rs[rsi--] ; /* An inline Return. Data instance words are CALLed,
				but do their own Return.  */
NEXT


 /*============================================= */

/* AND		( pytea pyteb ---  aANDb ) the bitwise Boolean		*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"AND\"
	.align	4, 0
	.int	0b
	.equ	ANDCFA, .
	");
AND:


 /* example partial optimization using GNU C compound lvalues and
	taking the address of an array (ds0), allowing accessing a
	byte array as ints. 
*/

DROP
if (Size == 4)
	{
	*(int *)(ds0 + dsi) =
		*(int *)(ds0 + dsi) & *(int *)(ds0 + dsi + 4);
	}
else	{
	for ( i = 0 ; i < Size	; i++ )
		{  j=dsi+i;
		*(byte *)(ds0 + j)  =
			*(byte *)(ds0 + j) & *(byte *)(ds0 + j +Size);
		}
	}
NEXT


 /*============================================= */

/* aint		( flagp --- !flagp ) casual NOT, bitwise on lsB.	*/
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"aint\"
	.align	4, 0
	.int	1b
	.equ	aintCFA, .
	");
aint:

ds[dsi] = ~ ds[dsi] ;
NEXT

 /*============================================= */

/* align		(P  a  --- a_rounded_up_to_CELLSIZE)		*/
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"align\"
	.align	4, 0
	.int	0b
	.equ	alignCFA, .
	");
align:


ps[psi] = (ps[psi] + 3) & 0xfffffffc ;
NEXT

 /*============================================= */

/* bump		( ---  old_pyte ) undrop. Only useful in threads. */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"bump\"
	.align	4, 0
	.int	1b
	.equ	bumpCFA, .
	");
bump:


dsi = dsi + Size ;
NEXT

 /* =============================== */

/* bytes		1 -> Size		*/
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"bytes\"
	.align	4, 0
	.int	0b
	.equ	bytesCFA, .
	");
bytes:

Size = 1 ;
NEXT

 /*============================================= */

/* bbytesd	bump 1->Size drop      */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"bbytesd\"
	.align	4, 0
	.int	1b
	.equ	bytes_bdCFA, .
	");
bytes_bd:

dsi = dsi + Size ;
Size = 1 ;
dsi = dsi - Size ;
NEXT


 /* =============================== */

/* bcellsd	bump sizeof(int)->Size drop	 */
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"bcellsd\"
	.align	4, 0
	.int	0b
	.equ	cells_bdCFA, .
	");
cells_bd:

dsi = dsi + Size ;
Size = CELLSIZE ;
dsi = dsi - Size ;
NEXT


 /* =============================== */

/* cells		sizeof(int) -> Size		*/
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"cells\"
	.align	4, 0
	.int	1b
	.equ	cellsCFA, .
	");
cells:

Size = CELLSIZE ;
NEXT


  /*============================================= */

/* drop		( pyte --- )			*/
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"drop\"
	.align	4, 0
	.int	0b
	.equ	dropCFA, .
	");
drop:

dsi = dsi - Size ;
NEXT


 /*============================================= */

/* dup		( pytea ||| ---	 pytea )  i.e.	  ( a --- a a )	   */
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"dup\"
	.align	4, 0
	.int	1b
	.equ	dupCFA, .
	");
dup:

BUMP
for (i = 0; i < Size; i++)
	ds[dsi+i]=ds[dsi+i-Size];
NEXT


 /*============================================= */

/* downshift	( shiftee amount --- shifted ) down-significance bitshift*/
  

asm	("
	1:
	.byte	9
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"downshift\"
	.align	4, 0
	.int	0b
	.equ	dshiftCFA, .
	");
dshift:

DROP
switch (Size){

case 4 :
	*(int *)(ds0+dsi) = *(int *)(ds0+dsi) >> *(char *)(ds0+dsi+Size);
	break;
case 1 :
	*(char *)(ds0+dsi) = *(char *)(ds0+dsi) >> *(char *)(ds0+dsi+Size);
	break;
case 2 :
	*(short *)(ds0+dsi) = *(short *)(ds0+dsi) >> *(char *)(ds0+dsi+Size);
	break;
case 3 :
	*(int *)(ds0+dsi) = *(int *)(ds0+dsi) >> *(char *)(ds0+dsi+Size);
	break;
default:
	*(int *)(ds0+dsi) = *(int *)(ds0+dsi) >> *(char *)(ds0+dsi+Size);
	break;
}
NEXT


 /*============================================= */

/* double		up-roll Size by 1, > 256 wraps	*/
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"double\"
	.align	4, 0
	.int	1b
	.equ	dubbleCFA, .
	");
dubble:


convert.i = Size << 1 ;
Size = (int)(convert.b[0] | convert.b[1]) ;
NEXT


 /*============================================= */

/* doHNC		(P &HNC	 --- ) (R; --- RET|null) execute word given &HNC */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"doHNC\"
	.align	4, 0
	.int	0b
	.equ	doHNCCFA, .
	");
doHNC:


 /* both cases below do an inline TOcode, but there's an extra degree of
	indirection with a thread CFA, which is a pointer to code.
	An atom's CFA *is* executable machine code.			*/

if ( !	(*(int *)ps[psi] & 0x00800000))	  /* if atomic bit not set....	*/
	{
	rs[++rsi] = (int )ip ;		/* push a return addy		*/

	/*  contentsof.. |<-&HNC->| + |<--- offset to CFA	     --->| */
	ip = (int *)	 (ps[psi]  +  ((*(byte *)ps[psi] + 3)  & 0xfc) + 8) ;

	/* pdrop */
	psi--;
	NEXT
	}

 /*	|<-&HNC->| + |<--- offset to CFA		-->| */
goto *	(ps[psi--]   + ((*(byte *)ps[psi] + 3 ) & 0xfc) + 8 ) ;


 /*============================================= */

/* emit		(pyte --- )  ascii print lsB. not buffered.    */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"emit\"
	.align	4, 0
	.int	1b
	.equ	emitCFA, .
	");
emit:

__asm__	 ("int $0x80"
	: "=a" (convert.i)
	: "0" (4),"b" (1),"c" (&ds[dsi]),
	  "d" (1) );
DROP
NEXT


 /*============================================= */

/* ell		unconditional branch to address in next thread cell */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ell\"
	.align	4, 0
	.int	0b
	.equ	ellCFA, .
	");
ell:

ip = (int *)*ip ;	  /*  put the int following this instr. in ip */
NEXT


 /*============================================= */

/* false		( --- 0flagpyte )  fast false lsB  */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"false\"
	.align	4, 0
	.int	1b
	.equ	falseCFA, .
	");
false:


BUMP
ds[dsi]= 0;
NEXT


 /*============================================= */

/* @		(P ptr ||| --- ) ( --- pyte ) fetch pyte at ptr at Size */
  

asm	("
	1:
	.byte	1
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"@\"
	.align	4, 0
	.int	0b
	.equ	fetchCFA, .
	");
fetch:


BUMP
for ( i = 0 ; i < Size ; i++)
	{ ds[dsi + i] = *(byte *)(ps[psi]+i) ;
	}
NEXT


 /*============================================= */

/* @+		(P ptr --- ptr+Size ) ( --- pyte )		 */
   

asm	("
	0:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"@+\"
	.align	4, 0
	.int	1b
	.equ	fetchplusCFA, .
	");
fetchplus:

				/* Bubb^H^H^H Chuck likes it	*/
BUMP
for ( i = 0 ; i < Size ; i++)
	{ ds[dsi + i] = *(byte *)(ps[psi]+i)  ;
	}
ps[psi] = ps[psi] + Size;
NEXT


 /*============================================= */

/* getSize	(P --- Size )	 */
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"getSize\"
	.align	4, 0
	.int	0b
	.equ	getSizeCFA, .
	");
getSize:

ps[++psi] = Size ;
NEXT


 /*============================================= */

/* gap		(P a b ||| ) ( --- a-b ) bytes between addresses */
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"gap\"
	.align	4, 0
	.int	1b
	.equ	gapCFA, .
	");
gap:

BUMP
convert.i =  ps[psi - 1] - ps[psi]  ;
for ( i = 0 ; i < CELLSIZE ; i++ )
	{ ds[dsi +i] = convert.b[i] ;
	}
NEXT

 /*============================================= */

/* halve		down-roll Size by 1 bit, i.e. Size=Size/2 usually*/
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"halve\"
	.align	4, 0
	.int	0b
	.equ	halveCFA, .
	");
halve:


convert.i = Size << 7 ;
Size = convert.b[0] | convert.b[1] ;
NEXT


 /*============================================= */

/* H3sm		no-op. No action besides increment VM instruction pointer */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"H3sm\"
	.align	4, 0
	.int	1b
	.equ	H3smCFA, .
	");
H3sm:

NEXT


 /*============================================= */

/* if		( pyte --- flagpyte ) OR pyte into lsB, make it a flag */
  

asm	("
	1:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"if\"
	.align	4, 0
	.int	0b
	.equ	flagCFA, .
	");
flag:

bite = 0 ;
for ( i= 0 ; i < Size ; i++ )
	{ if ( ds[dsi + i] )
		{ bite = 255 ;
		   break ;
		}
	}
ds[dsi] = bite ;
NEXT


 /*============================================= */

/* ifcontents=	(P &a &b ||| ) ( ---  flag ) compare pointed-to pytes	*/
   

asm	("
	0:
	.byte	11
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ifcontents=\"
	.align	4, 0
	.int	1b
	.equ	ifcontentsequalCFA, .
	");
ifcontentsequal:

			/* HAS been seen to produce 2 different flags */
BUMP
bite = 255 ;
for ( i = 0 ; i < Size ; i++ )
	{ ds[dsi + i] = 0;
	if (  *(byte *)(ps[psi-1]+i) != *(byte *)(ps[psi]+i)   )
		{ bite = 0 ;
		}
	}
ds[dsi] = bite ;
NEXT


 /*============================================= */

/* if=		( a b --- flagpyte ) 0xXX..ff (true) if a = b  */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"if=\"
	.align	4, 0
	.int	0b
	.equ	ifequalCFA, .
	");
ifequal:

			/* are top two pytes equal?   flag for ifbranch/tee */
bite = 255;			     /* boolean accum set to true */
DROP
for ( i = dsi ; i < dsi + Size ; i++ )
	{ if ( ds[i] != ds[i +	Size] )
		{  bite = 0;	    /* not =, set to false and exit loop */
		break;
		}
	}
ds[dsi] = bite;		      /* flagbyteTOS set to result boolean */
NEXT


 /*============================================= */

/* ifpositive	( a  --- flagpyte )  0 considered positive  */
   

asm	("
	0:
	.byte	10
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ifpositive\"
	.align	4, 0
	.int	1b
	.equ	ifpositiveCFA, .
	");
ifpositive:


if ((byte)ds[dsi+Size-1] & (byte)0x80 )
	{ ds[dsi] = 0 ;
	}
else	{ ds[dsi] = 0xff ;
	}
NEXT


 /*============================================= */

/* ip+		( ip_index --- ) execution-array mechanism, computed goto */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ip+\"
	.align	4, 0
	.int	0b
	.equ	ipplusCFA, .
	");
ipplus:

e = Size;
if ( Size > CELLSIZE ) e = CELLSIZE ;
convert.i = 0;
while (e)
	{ e--;
	convert.b[e] = ds[dsi+e];
	}
ip = (int *)*(ip + convert.i) ;
DROP
NEXT	/* CASE BRANCH */


  /* ============================================= */

/* literal	( --- pyte )  in-thread literal pyte number to ds
	works at Size 4 and 1 */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"literal\"
	.align	4, 0
	.int	1b
	.equ	literalCFA, .
	");
literal:


 /* put a pyte from the thread onto the data stack, runtime action of
a literal number in a thread. Literals have a size cell, but are pushed
at *current* Size. ip is a pointer to int, so we use convert.i/b.

This word should generalize to big pytes with an int loop if we clobber
the undrop area of the stack if litsize > Size, thus avoiding a different
big_literal word.
*/

j = *ip++ ;			/* j = lit size */
convert.i = *ip;		/* convert.i = lit value */
ip ++ ;				/* skip size cell and cell literal itself */
BUMP
ds[dsi	] =  convert.b[0];
ds[dsi+1] =  convert.b[1];
ds[dsi+2] =  convert.b[2];
ds[dsi+3] =  convert.b[3];
NEXT

  /*  something like...
j = (*ip++/CELLSIZE)+1;
i = 0;
while (j--)
	{ convert.i = *ip++;
	ds[dsi+i++] = convert.b[0]
	ds[dsi+i++] = convert.b[1]
	ds[dsi+i++] = convert.b[2]
	ds[dsi+i++] = convert.b[3]
	}
NEXT
*/


 /*============================================= */

/* max		( a b --- maxab ) Min's hubby.			*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"max\"
	.align	4, 0
	.int	0b
	.equ	maxCFA, .
	");
max:


					/* neg = 128, pos = 0 */
signA = ds[dsi + Size - 1] & 128 ;
DROP
signB = ds[dsi + Size - 1] & 128 ;

if ( signA > signB )			/* negative was dropped, done. */
	{ NEXT }
if ( signA < signB )			/* positive was dropped */
	{ MASH	NEXT }
i = Size;				/* if we're here signs are equal */
while (i--)		/* walk the bytes from msB till they differ */
	{ if ( ds[dsi + i] > ds[dsi + i + Size])
		{ NEXT }
	  if ( ds[dsi + i] < ds[dsi + i + Size])
		{ MASH NEXT
		}
	}	/* bytes equal at latest magnitude, try next lower byte */
NEXT		/* no bytes differ, numbers equal, finito. */


 /*============================================= */

/* min		( a b --- min_ab) Max's better half.		*/
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"min\"
	.align	4, 0
	.int	1b
	.equ	minCFA, .
	");
min:

					/* neg = 128, pos = 0		*/
signA = ds[dsi + Size - 1] & 128 ;
DROP
signB = ds[dsi + Size - 1] & 128 ;

if ( signA < signB )			/* negative was dropped, done.	*/
	{ NEXT }
if ( signA > signB )			/* positive was dropped		*/
	{ MASH	NEXT }
i = Size;				/* if we're here signs are equal */
while (i--)		/* walk the bytes from msB till they differ */
	{ if ( ds[dsi + i] < ds[dsi + i + Size])
		{ NEXT }
	  if ( ds[dsi + i] > ds[dsi + i + Size])
		{ MASH NEXT
		}
	}	/* bytes equal at latest magnitude, try next lower byte */
NEXT		/* no bytes differ, numbers equal, finito. */


 /*============================================= */

/* maskbyte	( a --- a&0x00...00ff ) mask off all but lsB   */
  

asm	("
	1:
	.byte	8
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"maskbyte\"
	.align	4, 0
	.int	0b
	.equ	maskbyteCFA, .
	");
maskbyte:

	       /* a violation of my at_least_3_component_words factoring rule.
			This is a just pyte constant and an AND, but it seems
			to make sense as one word, perhaps even in silicon.
			For one thing, this gets used a lot. */
ds[dsi] = ds[dsi] & 255 ;
for ( i = 1 ; i < Size ; i++ )
	{ ds[dsi +i] = 0 ;
	}
NEXT


 /*============================================= */

/* move		(P from to |||) (cnt ---)  move cnt bytes, overlaps are OK.
			You can't use this with the input buffer and 'this' */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"move\"
	.align	4, 0
	.int	1b
	.equ	moveCFA, .
	");
move:


convert.i = 0 ;
for ( i = 0 ; (i < CELLSIZE) & (i < Size) ; i++)
	{ convert.b[i] = ds[dsi+i];
	}
DROP
if ( ps[psi] > ps[psi -1] )
	{  for (i = convert.i-1 ; i >= 0 ; i--)
		{ *(byte *)(ps[psi]+i) = *(byte *)(ps[psi-1]+i) ;
		}
	}
else	{  for (i = 0 ; i < convert.i ; i++)
		{ *(byte *)(ps[psi -1]+i) = *(byte *)(ps[psi]+i) ;
		}
	}
NEXT


 /*============================================= */

/* -?r		( --- flagpyte )  (R c --- c-1) is TOR nonzero before decrement?*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"-?r\"
	.align	4, 0
	.int	0b
	.equ	minusaskrCFA, .
	");
minusaskr:


BUMP
if( rs[rsi]-- )
	{ *(char *)(ds0+dsi) = 255;
	}
else	{ *(char *)(ds0+dsi) = 0 ;
	}
NEXT


 /*============================================= */

/* NOT		( pyte --- !pyte )   bitwise NOT of all bits	*/
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"NOT\"
	.align	4, 0
	.int	1b
	.equ	NOTCFA, .
	");
NOT:


i = Size;
while (i-- )
	{ ds[dsi +i] = ~ ds[dsi +i];
	}
NEXT


 /*============================================= */

/* negate		( a --- 2's_complement_negative_a )	*/
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"negate\"
	.align	4, 0
	.int	0b
	.equ	negateCFA, .
	");
negate:

bite = 1 ;
for ( i = dsi ; i < dsi + Size ; i++ )
	{ if ( ds[i]==0 && bite ) bite |= 2 ;
	ds[i] = ((byte)0xff - ds[i]) + (bite&1) ;
	bite = bite>>1 ;
	}
NEXT


 /*============================================= */

/* no		( flagpyte --- ) branch to address in ip (the next cell) if 0  */
   

asm	("
	0:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"no\"
	.align	4, 0
	.int	1b
	.equ	noCFA, .
	");
no:

if ( ds[dsi] )		/* conditional on low-order byte of current pyte */
	{ ip++ ;		/* skip branch target addy */
	}
else
	{ ip = (int *)*ip;	/* put the int following this instr. in ip */
	}
DROP
NEXT


 /*============================================= */

/* over		( a b --- a b a ) the Forth usual OVER, but on pytes.	*/
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"over\"
	.align	4, 0
	.int	0b
	.equ	overCFA, .
	");
over:

BUMP
e = Size << 1 ;			      /* 2x Size */
for (i = 0  ; i < Size	; i++ )
	{ ds[dsi + i]=ds[dsi + i - e];
	}
NEXT


 /*============================================= */

/* onbits		( --- -1 )  all_ones pyte constant, 2's compl. -1 */
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"onbits\"
	.align	4, 0
	.int	1b
	.equ	onbitsCFA, .
	");
onbits:

BUMP
i = Size ;
while (i-- )
	{ ds[dsi + i ] = 255 ;
	}
NEXT

 /*============================================= */

/* one		( --- 1	 ) 1 as a pyte constant	  */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"one\"
	.align	4, 0
	.int	0b
	.equ	oneCFA, .
	");
one:

BUMP
ds[dsi] = 1 ;
for ( i = 1 ; i < Size ; i++ )
	{ ds[dsi + i] = 0 ;
	}
NEXT

 /*============================================= */

/* OR		( pytea pyteb --- pyteaORb ) bitwise Booolean OR    */
   

asm	("
	0:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"OR\"
	.align	4, 0
	.int	1b
	.equ	ORCFA, .
	");
OR:

DROP
i = Size ;
while (i-- )
      {	 j = dsi + i ;
	  ds[j] = ds[j + Size] | ds[j] ;
	}
NEXT


 /*============================================= */

/* pdrop		(P ptr --- )  decrement pointer stack index */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pdrop\"
	.align	4, 0
	.int	0b
	.equ	pdropCFA, .
	");
pdrop:

--psi;
NEXT


 /*============================================= */

/* p@		(P ptr1	 ---  ptr2 ) ptr1 overwritten */
   

asm	("
	0:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p@\"
	.align	4, 0
	.int	1b
	.equ	pfetchCFA, .
	");
pfetch:

				/* Forth @, if param stack is ptr stack */
ps[psi] = *(int *)ps[psi]   ;
NEXT


 /*============================================= */

/* pdup		(P ptra --- ptra ptra )	 */
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pdup\"
	.align	4, 0
	.int	0b
	.equ	pdupCFA, .
	");
pdup:

ps[++psi] = ps[psi -1 ] ;
NEXT


 /*============================================= */

/* pliteral	(P --- lit )	next cell in thread to P.
		pointless from interpreter, needs a threading wrapper. */
   

asm	("
	0:
	.byte	8
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pliteral\"
	.align	4, 0
	.int	1b
	.equ	pliteralCFA, .
	");
pliteral:


ps[++psi] = *ip++ ;
NEXT


 /*============================================= */

/* +		( a b --- a+b )	 2's-complement addition of pytes  */
  

asm	("
	1:
	.byte	1
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"+\"
	.align	4, 0
	.int	0b
	.equ	plusCFA, .
	");
plus:

DROP
bite = 0 ;
for ( i = dsi ; i < dsi + Size	; i++ )
	{ if ( (~ds[i]&0xff)<ds[i+Size] || ((~ds[i]&0xff)==ds[i+Size]&&bite) )
 /* COMMENT THIS. OPTIMIZE THIS.  */
	     bite |= 2 ;
	ds[i] += ds[i+Size] + (bite&1) ; bite = bite>>1 ;
	}
NEXT


 /*============================================= */

/* pover		pea over  (P a b --- a b a ) pointer-stack OVER */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pover\"
	.align	4, 0
	.int	1b
	.equ	poverCFA, .
	");
pover:

ps[++psi] = ps[psi -2 ] ;
NEXT


 /*============================================= */

/* p+		( pyte --- ) (P a --- a+(int_part_or_less_of)pyte ) */
  

asm	("
	1:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p+\"
	.align	4, 0
	.int	0b
	.equ	pluspCFA, .
	");
plusp:

convert.i = 0;
j = Size;
if ( Size > CELLSIZE )
	j = CELLSIZE;
for (i=0 ; i < j ; i++ )
	{ convert.b[i] = ds[dsi+i];
	}
ps[psi] = ps[psi] + convert.i ;
DROP
NEXT


 /*============================================= */

/* p+s		(P a --- a+Size ) index a pointer by Size	*/
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p+s\"
	.align	4, 0
	.int	1b
	.equ	pplussCFA, .
	");
ppluss:

ps[psi] = ps[psi] + Size;
NEXT


 /*============================================= */

/* p+1		(P a --- a+1 )	index a pointer by one		*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p+1\"
	.align	4, 0
	.int	0b
	.equ	pplusbCFA, .
	");
pplusb:

ps[psi] = ps[psi] + 1 ;
NEXT


 /*============================================= */

/* p+c		(P a --- a+cellsize ) index a pointer by cellsize (x86 is 4) */
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p+c\"
	.align	4, 0
	.int	1b
	.equ	ppluscCFA, .
	");
pplusc:

ps[psi] = ps[psi] + CELLSIZE ;
NEXT


 /*============================================= */

/* p-c		(P a --- a-cellsize ) decrement pointer by cellsize (4 on x86) */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p-c\"
	.align	4, 0
	.int	0b
	.equ	pminuscCFA, .
	");
pminusc:

ps[psi] = ps[psi] - CELLSIZE ;
NEXT


 /*============================================= */

/* p-1		(P ptr --- ptr+1 ) decrement a pointer			*/
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p-1\"
	.align	4, 0
	.int	1b
	.equ	pminus1CFA, .
	");
pminus1:

ps[psi] = ps[psi] - 1 ;
NEXT


 /*============================================= */

/* p->		(P a --- ) ( --- a_low[a_hi] ) if (Size > cellsize) then 0-pad	*/

  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p->\"
	.align	4, 0
	.int	0b
	.equ	pTOCFA, .
	");
pTO:

BUMP
for ( i = CELLSIZE ; i < Size ; i++ )	/* do an in-line zero first maybe */
	{	ds[dsi + i] = 0;
	}
*(int *)(ds0+dsi) = ps[psi--] ;
NEXT


 /*============================================= */

/* p->r		(P a --- ) (R; --- a ) */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p->r\"
	.align	4, 0
	.int	1b
	.equ	pTOrCFA, .
	");
pTOr:

rs[++rsi] = ps[psi--] ;
NEXT


 /*============================================= */

/* p->S		(P Size --- )  !Size   superceeded?	*/
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p->s\"
	.align	4, 0
	.int	0b
	.equ	pTOsCFA, .
	");
pTOs:

Size = ps[psi]	& 0x1ff;	/* 19991225 */
psi--;
NEXT


 /*============================================= */

/* p!		pea store (P p  store	 |||  ) p goes in store, no pdrops  */
   

asm	("
	0:
	.byte	2
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"p!\"
	.align	4, 0
	.int	1b
	.equ	pstoreCFA, .
	");
pstore:


*(int*)ps[psi] = ps[psi-1] ;
NEXT


 /*============================================= */

/* pswap		(P a b --- b a )	*/
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pswap\"
	.align	4, 0
	.int	0b
	.equ	pswapCFA, .
	");
pswap:

scratch = ps[psi] ;
ps[psi] = ps[psi - 1] ;
ps[psi - 1] = scratch ;
NEXT


 /*============================================= */

/* pup		(P ---	oldptr )		       */
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pup\"
	.align	4, 0
	.int	1b
	.equ	pupCFA, .
	");
pup:

++psi;
NEXT


 /*============================================= */

/* return		(R; xt	--- )  VMST rts to calling thread   */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"return\"
	.align	4, 0
	.int	0b
	.equ	ReturnCFA, .
	");
Return:

ip = (int *)rs[rsi--];
NEXT				/* goto **ip++; */


 /*============================================= */

/* rdrop		(R; a --- ) */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"rdrop\"
	.align	4, 0
	.int	1b
	.equ	rdropCFA, .
	");
rdrop:

rsi-- ;
NEXT


 /*============================================= */

/* rpcopy		(R; a ||| ) (P --- a ) dup r to p   */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"rpcopy\"
	.align	4, 0
	.int	0b
	.equ	rpcopyCFA, .
	");
rpcopy:

ps[++psi] = rs[rsi] ;
NEXT


 /*============================================= */

 /* r-1		(R: val --- val-1 )  decrement TOR
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"r-1\"
	.align	4, 0
	.int	1b
	.equ	rminus1CFA, .
	");
rminus1:

rs[rsi]--;
NEXT */


 /*============================================= */

/* <-r		(R; val --- ) ( --- 0000val )  4 live bytes written to ds  */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"<-r\"
	.align	4, 0
	.int	0b
	.equ	rTOCFA, .
	");
rTO:

BUMP
for ( i = CELLSIZE ; i < Size ; i++ )	/* zero out bytes > CELLSIZE  */
	{	ds[dsi + i] = 0;
	}
*(int *)(ds0+dsi) = rs[rsi--] ;
NEXT


 /*============================================= */

/* r->p		(R; ptr --- ) (P --- ptr ) symmetrical with p->r. */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"r->p\"
	.align	4, 0
	.int	1b
	.equ	rTOpCFA, .
	");
rTOp:

ps[++psi] =  rs[rsi--] ;
NEXT


 /*============================================= */

/* r->s		(R; size --- )	!SIZE */
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"r->s\"
	.align	4, 0
	.int	0b
	.equ	rTOsCFA, .
	");
rTOs:

Size = rs[rsi--] & 0x1ff ;	/* 19991225 */
NEXT


 /* ============================================= */

/* swap		( pytea pyteb --- pyteb pytea ) */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"swap\"
	.align	4, 0
	.int	1b
	.equ	swapCFA, .
	");
swap:


j = dsi - Size ;
for ( i = 0; i < Size  ; i++ )
	{ bite = ds[dsi + i] ;
	ds[dsi + i] = ds[dsi -Size + i] ;
	ds[dsi - Size  + i] = bite ;
	}
NEXT


 /*============================================= */

/* setSize	( Size --- )  Silently refuses to set Size to 0. */
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"setSize\"
	.align	4, 0
	.int	0b
	.equ	setSizeCFA, .
	");
setSize:

if ( ds[dsi] & 0xff )
Size = ds[dsi] ;	/* 19991225 */
DROP
NEXT


 /*============================================= */

/* s->r		(R; --- Size ) Size to return stack		*/
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"s->r\"
	.align	4, 0
	.int	1b
	.equ	sTOrCFA, .
	");
sTOr:

rs[++rsi] = Size ;
NEXT


 /*============================================= */

/* s->p		(P --- Size )  Size to pointer stack		*/
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"s->p\"
	.align	4, 0
	.int	0b
	.equ	sTOpCFA, .
	");
sTOp:

ps[++psi] = Size ;
NEXT


 /*============================================= */
					/* Got milk? */

/* !		(P ptr ||| ) ( pyte --- ) pointer is not dropped */
   

asm	("
	0:
	.byte	1
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"!\"
	.align	4, 0
	.int	1b
	.equ	storeCFA, .
	");
store:

 /* OPTIMIZE ME. Why? p! is fast. */
for ( i = 0 ; i < Size ; i++ )
	{ *(byte *)(ps[psi]+i) = ds[dsi+i];
	}
DROP
NEXT


 /* ============================================= */

/* sub		(R; --- ip+ ) ip!   go subthread, other uses. Essence of VMST	*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"sub\"
	.align	4, 0
	.int	0b
	.equ	subCFA, .
	");
sub:

				/* recall that ip has been incremented by
					NEXT already */
rs[++rsi] = (int)(ip + 1 ) ;	/* push cell after callee cell to rs  */
ip = (int *)*ip;		/* ip is (callee cell), NEXT will jsr */
NEXT				/*  goto **ip++ */


 /* ============================================= */

/* ->r		(R; --- val ) ( val --- ) Forth >r, ish. 4 bytes moved. */
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"->r\"
	.align	4, 0
	.int	1b
	.equ	TOrCFA, .
	");
TOr:

rs[++rsi] = *(int *)(ds0+dsi) ;
DROP
NEXT


 /*============================================= */

/* ->code		(P &HNC --- Code_Body_Cell ) */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"->code\"
	.align	4, 0
	.int	0b
	.equ	TOcodeCFA, .
	");
TOcode:

ps[psi] = ps[psi] +  ((*(byte *)ps[psi] + 3)  & 0xfc) + CELLSIZE + CELLSIZE  ;
NEXT


 /*============================================= */

/* ->link		(P &HNC --- Link_Cell ) */
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"->link\"
	.align	4, 0
	.int	1b
	.equ	TOlinkCFA, .
	");
TOlink:


ps[psi] = ps[psi] + ((*(byte *)ps[psi] + 3)  & 0xfc ) + CELLSIZE ;
NEXT


 /*============================================= */

/* ->p		(P --- CELL ) ( a --- ) CELLSIZE worth of a is moved  */
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"->p\"
	.align	4, 0
	.int	0b
	.equ	TOpCFA, .
	");
TOp:


ps[++psi]= *(int *)(ds0 + dsi );
DROP
NEXT


 /*============================================= */

/* true		( --- true_flagpyte ) 0xXXX...XXff, true flag pyte constant  */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"true\"
	.align	4, 0
	.int	1b
	.equ	trueCFA, .
	");
true:

BUMP
ds[dsi] = 0xff ;
NEXT


 /*============================================= */

/* 2pdrop		(P  A B --- )  drop 2 pointers. */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"2pdrop\"
	.align	4, 0
	.int	0b
	.equ	twopdropCFA, .
	");
twopdrop:

psi-- ;
psi-- ;			/* this one compiles to one instruction as is.
				So leave it. */
NEXT

 /*============================================= */

/* upshift	( shiftee amount ---  up-shifted-lo [up-shifted-high] ) */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"upshift\"
	.align	4, 0
	.int	1b
	.equ	upshiftCFA, .
	");
upshift:


cells = ds[dsi] >> 3 ;			/* cellshift value */
net = ds[dsi] - (cells << 3);
for ( i = 0 ; i < cells ; i++ )
	{ accum[i]=0;			/* leftpad 0-cells */
	}				/* otherwise there will be junk */
DROP					/* dsi is shiftee */
for ( i = 0 ; i < Size ; i++ )		/* cell&bit  to accum */
	{ accum[cells + i] = ds[dsi + i] << net ;
	}
ds[dsi] = accum[0]  ;			/* avoid accum underrun	 below */
for ( i = 1 ; i < Size ; i++ )
	{ ds[dsi + i] =
	    (accum[i] & 0x00ff ) | ((accum[i-1] & 0xff00)>> 8 ) ;
	}
NEXT

  /*============================================= */

/* x		( pyte --- ) print & drop a pyte, Forth . but hex only	  */
  

asm	("
	1:
	.byte	1
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"x\"
	.align	4, 0
	.int	0b
	.equ	printhexCFA, .
	");
printhex:

i = 0;
bite = 32;	/* ascii space */

	/*				build number string */
while(i < Size )
{
j= dsi + Size - 1  - i ;
	numberString[i << 1]	   = digitOut[	ds[j] >> 4  ];
	numberString[(i << 1)+ 1 ] = digitOut[	ds[j] & 15  ];
	i++;
}

				 /* print number string */
__asm__	 ("int $0x80"
	: "=a" (convert.i)
	: "0" (4),"b" (1),"c" (numberString),
	  "d" (Size << 1 ) );

				 /* print space		*/
__asm__	 ("int $0x80"
	: "=a" (convert.i)
	: "0" (4),"b" (1),"c" (&bite),
	  "d" (1) );
DROP
NEXT


  /*============================================= */

/* XOR		( pytea pyteb --- pyteaXORb )		*/
   

asm	("
	0:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"XOR\"
	.align	4, 0
	.int	1b
	.equ	XORCFA, .
	");
XOR:

DROP
for ( i = 0 ; i < Size ; i++ )
	ds[dsi + i]=ds[dsi + Size + i] ^ ds[dsi+i];
NEXT


 /*=============================================		*/

/* yes		( flagpyte --- )  branch if lsB of TOS is true	*/
  

asm	("
	1:
	.byte	3
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"yes\"
	.align	4, 0
	.int	0b
	.equ	yesCFA, .
	");
yes:

if ( ds[dsi] )		   /* conditional on low-order byte of current pyte */
	{ ip = (int *)*ip ;	   /* put the int following this instr. in ip */
	}
else	{ ip++ ;		  /* skip branch target addy */
	}
DROP
NEXT


 /*============================================= */

/* zero		( --- 0 )   0 as a fast pyte-constant  */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"zero\"
	.align	4, 0
	.int	1b
	.equ	zeroCFA, .
	");
zero:


BUMP
for ( i = 0 ; i < Size ; i++ )
	{	ds[dsi + i] = 0;
	}
NEXT

	/* cheap docs	*/

/* */
/*	SYSCALLS ...

see Linux asm/unistd.h, man syscalls, man 2 intro

No errno handling; Just one return value. errno is a derived value from
the natural low-level return value anyway, and you have H3m sitting in
front of you to play with it.

No C subroutine wrappers, so no name scoping. args
(inputs) are C expressions ( see gcc.info ).

Syscalls slam ints on and off the data data stack. This is noticably
smaller than using convert.i and/or testing Size, and I want to get a lot
of syscalls in here for cheap.	*/

 /* ==========sssssssssssssssssssssssssssssssss==== */


  

asm	("
	1:
	.byte	11
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"<<<SYSCALLS\"
	.align	4, 0
	.int	0b
	.equ	thxtebhtxxioCFA, .
	");
thxtebhtxxio:




/* Sbecome	(P  path argv envp ||| ) ( --- GONE|fail=-1) unix execve  */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"Sbecome\"
	.align	4, 0
	.int	1b
	.equ	SbecomeCFA, .
	");
Sbecome:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (11),"b" ((char *)(ps[psi-2])),"c" ((char *)(ps[psi-1])),
	  "d" ((char *)(ps[psi]))  );
NEXT
	/* only returns on failure	*/


/* close		( FD --- ret )	lose and release a file descriptor */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"close\"
	.align	4, 0
	.int	0b
	.equ	closeCFA, .
	");
close:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (6),"b" (*(int *)(ds0+dsi))  );
NEXT


/* Schroot	(P  path |||) ( --- ret) change apparent root directory */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"Schroot\"
	.align	4, 0
	.int	1b
	.equ	SchrootCFA, .
	");
Schroot:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (61),"b" ((char *)(ps[psi]))  );
NEXT


/* dupFD		(old_FD --- new_FID|-1)	  unix dup()	*/
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"dupFD\"
	.align	4, 0
	.int	0b
	.equ	dupFDCFA, .
	");
dupFD:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (41),"b" (*(int *)(ds0+dsi)) );
NEXT


/* dirents	( FD  count --- red|-1|0 )  (P	buff ||| ) unix getdents()  */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"dirents\"
	.align	4, 0
	.int	1b
	.equ	direntsCFA, .
	");
dirents:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (141),"b" ((char *)(ps[psi])),"c" (*(int *)(ds0+dsi+Size)),
	  "d" (*(int *)(ds0+dsi))  );
NEXT


/* FDcontrol	( FD CMD  --- ret )  (P [arg] |||)  unix fcntl()
		multimode, more args given than used, ordered so
		that options are lower on stack	 */
  

asm	("
	1:
	.byte	9
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"FDcontrol\"
	.align	4, 0
	.int	0b
	.equ	FDcontrolCFA, .
	");
FDcontrol:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (55),"b" (*(int *)(ds0+dsi+Size)),
	  "c" (*(int *)(ds0+dsi)),
	  "d" ((char *)(ps[psi]))  );
NEXT


/* fork		parent.. ( ---PID_of_child_or-1)	child.. (--- 0)*/
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"fork\"
	.align	4, 0
	.int	1b
	.equ	forkCFA, .
	");
fork:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (2)  );
NEXT


/* fflush		( FD --- flag )	 unix fsync()		*/
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"fflush\"
	.align	4, 0
	.int	0b
	.equ	fflushCFA, .
	");
fflush:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (118), "b" (*(int *)(ds0+dsi))  );
NEXT


/* flush		( --- 0 )	unix sync()	*/
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"flush\"
	.align	4, 0
	.int	1b
	.equ	flushCFA, .
	");
flush:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (36)  );
NEXT


/* Sgettimeofday	(P timeval |||) ( --- flag)  simplified */
  

asm	("
	1:
	.byte	13
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"Sgettimeofday\"
	.align	4, 0
	.int	0b
	.equ	SgettimeofdayCFA, .
	");
Sgettimeofday:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (78),"b" (*(int *)(ps[psi])),"c" (0)  );
NEXT


/* link		(P  old new ||| )   ( --- flag )   unix link()	*/
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"link\"
	.align	4, 0
	.int	1b
	.equ	linkCFA, .
	");
link:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (9),"b" ((char *)(ps[psi])),
		  "c" ((char *)(ps[psi-1]))  );
NEXT


/* ioctl		( FD  ioctl --- ret) (P	 argp --- ) note the pdrop */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ioctl\"
	.align	4, 0
	.int	0b
	.equ	ioctlCFA, .
	");
ioctl:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (54),"b" (*(int *)(ds0+dsi+Size)),
	  "c" (*(int *)(ds0+dsi)),
	  "d" ((char *)(ps[psi]))  );
psi--;
NEXT


/* ioperm		( level --- flag )	Linux iopl()	*/
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"ioperm\"
	.align	4, 0
	.int	1b
	.equ	iopermCFA, .
	");
ioperm:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (110), "b" (*(int *)(ds0+dsi))  );
NEXT


/* makedir	(P name ||| )	( --- flag)	unix mkdir(), perms==0	*/
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"makedir\"
	.align	4, 0
	.int	0b
	.equ	makedirCFA, .
	");
makedir:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (39),"b" ((char *)(ps[psi])),"c" (0) );
NEXT


/* mount		(P special dir fstype|NULL data |||) (rwflags --- flag) */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"mount\"
	.align	4, 0
	.int	1b
	.equ	mountCFA, .
	");
mount:

i = ps[psi-3];			/* special file		*/
j = ps[psi-2];			/* mountpoint		*/
k = ps[psi-1];			/* fstype|NULL		*/
K = ps[psi];			/* r/w flags		*/
e = *(int *)(ds0 + dsi);	/* data			*/
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (21),"b" (i),"c" (j),"d" (k),"D" (e),
		"S" (K )  );
NEXT


/* nanosleep	(P  timespec remaining ||| )	( --- flag )  */
  

asm	("
	1:
	.byte	9
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"nanosleep\"
	.align	4, 0
	.int	0b
	.equ	nanosleepCFA, .
	");
nanosleep:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (162),"b" ((char *)(ps[psi-1])),
		    "c" ((char *)(ps[psi]))  );
NEXT


/* owner		(P path |||)   ( UID  GID --- ret)  unix chown() */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"owner\"
	.align	4, 0
	.int	1b
	.equ	ownerCFA, .
	");
owner:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (182),"b" ((char *)(ps[psi])),
	  "c" (*(int *)(ds0+dsi+Size)),
	  "d" (*(int *)(ds0+dsi))  );
NEXT


/* permit		(P path |||)   ( mode --- ret)	 unix chmod()	*/
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"permit\"
	.align	4, 0
	.int	0b
	.equ	permitCFA, .
	");
permit:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (15),"b" ((char *)(ps[psi])),
		   "c" (*(int *)(ds0+dsi)) );
NEXT


/* pause		( --- -1 )	wait for a signal */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pause\"
	.align	4, 0
	.int	1b
	.equ	pauseCFA, .
	");
pause:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (29) );
NEXT


/* pipe		(P ptr_to_int[2] |||)  ( --- flag) 2 FD's in given array */
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"pipe\"
	.align	4, 0
	.int	0b
	.equ	pipeCFA, .
	");
pipe:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (42),"b" ((char *)(ps[psi])) );
NEXT
 /* do something handy with ptr_to_int[2] */


/* parent		( --- parent_PID )	unix getppid()		*/
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"parent\"
	.align	4, 0
	.int	1b
	.equ	parentCFA, .
	");
parent:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (64)	);
NEXT


/* quit		unix exit(), Forth bye, H3m quit. q. works. */
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"quit\"
	.align	4, 0
	.int	0b
	.equ	quitCFA, .
	");
quit:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (1),"b" (ds[dsi]) );
NEXT
 /* NEXT would never happen, but it seems to be needed here to keep the
	dictionary linking wholesome or something.
	DISTURBING. It does seem to be limited to exit though, which is
		semi-OK.  */


/* read		( FD --- red ) (P buf ||| )  always requests 2048 bytes	 */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"read\"
	.align	4, 0
	.int	1b
	.equ	readCFA, .
	");
read:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (3),"b" (*(int *)(ds0+dsi)),"c" ((char *)(ps[psi])),
	  "d" (2048) );
NEXT


/* rename		(P old new |||) ( --- flag ) basis of mv     */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"rename\"
	.align	4, 0
	.int	0b
	.equ	renameCFA, .
	");
rename:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (38),"b" ( (char *)(ps[psi-1]) ),
		   "c" ( (char *)(ps[psi]) )
	);
NEXT


/* resize		(P end_data_segment ) ( --- ret )   unix brk(), basis of malloc */
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"resize\"
	.align	4, 0
	.int	1b
	.equ	resizeCFA, .
	");
resize:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (45),"b" ((char *)(ps[psi]))  );
NEXT


/* reboot		(P btparm --*POW!!* REBOOT LINUX. With EXTREME prejudice. */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"reboot\"
	.align	4, 0
	.int	0b
	.equ	rebootCFA, .
	");
reboot:

i = ps[psi];		/* address of bootparams string */

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (88), "b" (0xfee1dead), "c" (672274793),"d" ( 0xa1b2c3d4 ),
		"D" (i)	    );
NEXT


/* sys_open	(P name |||)  ( flags perms --- FD ) see open thread  */
   

asm	("
	0:
	.byte	8
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"sys_open\"
	.align	4, 0
	.int	1b
	.equ	sys_openCFA, .
	");
sys_open:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (5),"b" ((char *)(ps[psi])),
	  "c" (*(int *)(ds0+dsi+Size)),
	  "d" (*(int *)(ds0+dsi))  );
NEXT


/* todir		(P path --- ) (--- ret)	 unix chdir(), cd  */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"todir\"
	.align	4, 0
	.int	0b
	.equ	todirCFA, .
	");
todir:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (12),"b" ((char *)(ps[psi]))  );
NEXT


/* rmdir		(P name ||| )	( --- flag )  remove empty dir	*/
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"rmdir\"
	.align	4, 0
	.int	1b
	.equ	rmdirCFA, .
	");
rmdir:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (40), "b" ( (char *)(ps[psi]) ) );
NEXT


/* readlink	(P name buff ||| ) ( bufsize --- flag )	 unix readlink	*/
  

asm	("
	1:
	.byte	8
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"readlink\"
	.align	4, 0
	.int	0b
	.equ	readlinkCFA, .
	");
readlink:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (85) ,"b" ((char *)(ps[psi-1])),"c" ((char *)(ps[psi])),
		"d" ( *(int *)(ds0+dsi) )  );
NEXT


/* stat		(P name buff ||| ) ( --- flag)	get a file's stat data */
   

asm	("
	0:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"stat\"
	.align	4, 0
	.int	1b
	.equ	statCFA, .
	");
stat:


BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (106),"b" (*(int *)(ps[psi-1])),
		"c" (*(int *)(ps[psi])) );
NEXT


/* seek		( FD o/s whence --- o/s0 ) unix lseek(), Forth reposition-file */
  

asm	("
	1:
	.byte	4
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"seek\"
	.align	4, 0
	.int	0b
	.equ	seekCFA, .
	");
seek:

DROP DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (19),"b" (*(int *)(ds0+dsi)),
		   "c" (*(int *)(ds0+dsi+Size)),
		   "d" ( *(int *)(ds0+dsi+Size+Size) )	      );
NEXT


/* settimeofday	(P timeval |||) ( --- flag)   simplified	*/
   

asm	("
	0:
	.byte	12
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"settimeofday\"
	.align	4, 0
	.int	1b
	.equ	settimeofdayCFA, .
	");
settimeofday:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (79),"b" (*(int *)(ps[psi])),"c" (0)  );
NEXT


/* sigaction	( sig# --- flag ) (P action old |||) install a signal handler */
  

asm	("
	1:
	.byte	9
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"sigaction\"
	.align	4, 0
	.int	0b
	.equ	sigactionCFA, .
	");
sigaction:


i = (ps[psi-1]);
j = (ps[psi]);

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (67),"b" (* (int*)(ds0+dsi) ),
		   "c" (i),
		   "d" (j) );
NEXT


/* socketcall	( call --- flag )  (P args ||| )  Linux socket calls entry */
   

asm	("
	0:
	.byte	10
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"socketcall\"
	.align	4, 0
	.int	1b
	.equ	socketcallCFA, .
	");
socketcall:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (102), "b" (*(int *)(ds0+dsi)),
		     "c" (ps[psi])		);
NEXT


/* statfs		(P name buff ||| ) ( --- flag)	get a filesys stat data */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"statfs\"
	.align	4, 0
	.int	0b
	.equ	statfsCFA, .
	");
statfs:


BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (99),"b" (*(int *)(ps[psi-1])),
		"c" (*(int *)(ps[psi])) );
NEXT


/* settime	(P timestruct ||| )   ( --- flag) set sec. elapsed in epoch */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"settime\"
	.align	4, 0
	.int	1b
	.equ	stimeCFA, .
	");
stime:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (25),"b" (*(int *)(ps[psi]))  );
NEXT


/* swapoff	(P name ||| )	( --- flag )	*/
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"swapoff\"
	.align	4, 0
	.int	0b
	.equ	swapoffCFA, .
	");
swapoff:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (115),"b" (*(int *)(ps[psi]))  );
NEXT


/* swapon		(P name ||| )	( --- flag )	*/
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"swapon\"
	.align	4, 0
	.int	1b
	.equ	swaponCFA, .
	");
swapon:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (87),"b" (*(int *)(ps[psi])), "c" (0)  );
NEXT


/* special	(P name |||)	( mode dev --- flag )	unix mknod()	*/
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"special\"
	.align	4, 0
	.int	0b
	.equ	specialCFA, .
	");
special:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (14),"b" (*(int *)(ps[psi])),"c" (*(int *)(ds0+dsi)),
		"d" (*(int *)(ds0+dsi+Size)) );
NEXT


/* signal		( PID signal --- ret)	unix kill	*/
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"signal\"
	.align	4, 0
	.int	1b
	.equ	signalCFA, .
	");
signal:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (37),"b" (*(int *)(ds0+dsi+Size)),"c" (0),
	  "d" (*(int *)(ds0+dsi)) );
NEXT


/* symlink	(P exists new |||)  ( --- flag )	*/
  

asm	("
	1:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"symlink\"
	.align	4, 0
	.int	0b
	.equ	symlinkCFA, .
	");
symlink:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (83),"b" (*(int *)(ps[psi-1])),
		"c" (*(int *)(ps[psi])) );
NEXT


/* syslog		( type	len --- flag )	(P buf ||| ) read/tweak kbuffer */
   

asm	("
	0:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"syslog\"
	.align	4, 0
	.int	1b
	.equ	syslogCFA, .
	");
syslog:


DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (103),"b" (*(int *)(ds0+dsi)),"c" ((char *)(ps[psi])),
		"d" (*(int *)(ds0+dsi+Size)) );
NEXT


/* truncate	(P fname || )  ( len --- flag ) truncate fname to len  */
  

asm	("
	1:
	.byte	8
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"truncate\"
	.align	4, 0
	.int	0b
	.equ	truncateCFA, .
	");
truncate:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (92), "b" ((char *)(ps[psi])), "c" (*(int *)(ds0+dsi)) );
NEXT


/* trace		( request PID addr data --- flag)  unix ptrace() */
   

asm	("
	0:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"trace\"
	.align	4, 0
	.int	1b
	.equ	traceCFA, .
	");
trace:

DROP	DROP	DROP
i = *(int *)(ds0+dsi+Size+Size+Size);
j = *(int *)(ds0+dsi+Size+Size);
k = *(int *)(ds0+dsi+Size);
e = *(int *)(ds0+dsi);
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (26),"b" (i),"c" (j),
	  "d" (k),"D" (e) );
NEXT


/* unlink		(P name ||| ) ( --- flag) rm a name, maybe the data
			inode */
			/* a hardlink is a name of an inode. unlink
				the last one and unix frees the inode  */
  

asm	("
	1:
	.byte	6
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"unlink\"
	.align	4, 0
	.int	0b
	.equ	unlinkCFA, .
	");
unlink:

BUMP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (10), "b" ((char *)(ps[psi])) );
NEXT


/* unmount	(P name ||| )	( --- flag) unix umount() */
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	1
	.ascii	\"unmount\"
	.align	4, 0
	.int	1b
	.equ	unmountCFA, .
	");
unmount:

__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (22), "b" ((char *)(ps[psi])) );
NEXT


/* write		( FDbyte  request --- red )   (P name ||| )  */
  

asm	("
	1:
	.byte	5
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"write\"
	.align	4, 0
	.int	0b
	.equ	writeCFA, .
	");
write:

DROP
__asm__	 ("int $0x80"
 /* return */	: "=a" ( *(int *)(ds0+dsi)  )
 /* call # */	: "0" (4),
 /* FD */	"b" ( *(int *)(ds0+dsi) ),
 /* buffer */	"c" ( (char *)(ps[psi]) ),
 /* size */	"d" ( *(int *)(ds0+dsi+Size) )	 );
NEXT


/* waitpid	(pid options --- ret )	simplified	*/
   

asm	("
	0:
	.byte	7
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"waitpid\"
	.align	4, 0
	.int	1b
	.equ	waitpidCFA, .
	");
waitpid:

DROP
__asm__	 ("int $0x80"
	: "=a" (*(int *)(ds0+dsi))
	: "0" (7),"b" (*(int *)(ds0+dsi+Size)),"c" (0),
	  "d" (*(int *)(ds0+dsi)) );
NEXT



  

asm	("
	1:
	.byte	10
	.byte	0
	.byte	0x80
	.byte	0
	.ascii	\"<<<THREADS\"
	.align	4, 0
	.int	0b
	.equ	thxbtixioCFA, .
	");
thxbtixio:


 /* 7777777777777777777777777777777777777777777777777777 */
 /* end of dictionary build section for fundamentals/atoms/primitives.
Begin derived words; threads, named noun instances. The above is the
virtual machine, The non-syscall atoms are the virtual CPU.  the following
are some things you can do with said machine. Compiled-in threads are
written in the assembly language of the VM. Note that some of this
proto-H3m code is very un-Forth-like, doing many things wrong. That's
because H3m doesn't really exist yet, just the VM. So, it's nasty
tortuous virtual assembly. As far as examples of using the H3m
interpreter, Note the opcodes used, but ignore the coding style. Very
important in Virtual Machine Subroutine Threading, above here (by default)
words have the atomic bit set in the Head Name Interface Cell.
m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4m4

m4 macros for all the standard build_a_thread stuff.		*/

 /* contents of this int is given by  arg			*/


 /* this cell contains CFA of word named as argument		*/


 /* VMST jsr, takes a whole cell, plus the callee's cell
This doesn't have the asm	(" .int CFA ");
 folded into it to keep branch offset counting
1:1, except for branch targets. ___o___				*/


 /* build a relative branch offset with integer arg,
$1 is +/-, $2 is #   +/- needs a comma */


 /* we are at a branch target, set label symbols for here.
targets take no thread space themselves.
The graphicoid name is to emphasize that a cell is not created.	 */


 /* back-branch address of labels. arg-less macro. */


/* */
/*	NAMED DATA... */
 /* ddddddddddddddddddddddddddddddd============			*/

/* block		(P --- begin_buffer) object data buffer(s) */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"block\"
	.align 4, 0
	.int  1b
	.equ	blockCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 1024 ");


/* buffer		(P --- begin_buffer) "file" input buffer */
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"buffer\"
	.align 4, 0
	.int  0b
	.equ	bufferCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 2048 ");


/* bsi		counter for branches to resolve		*/
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 1
	.ascii \"bsi\"
	.align 4, 0
	.int  1b
	.equ	bsiCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");



/* branchstack	space for 16 branches to resolve	*/
  
asm	("
	1:
	.byte 11
	.byte 0
	.byte 0
	.byte 1
	.ascii \"branchstack\"
	.align 4, 0
	.int  0b
	.equ	branchstackCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 128 ");
	/* allot for a maximal pyte */


/* digits		(P  ---	 address_of_array ) ascii vals to hex vals map	*/
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"digits\"
	.align 4, 0
	.int  1b
	.equ	digitsCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 
asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm ("		.byte 0x00
		.byte 0x01
		.byte 0x02
		.byte 0x03	"); 
asm ("		.byte 0x04
		.byte 0x05
		.byte 0x06
		.byte 0x07	"); 
asm ("		.byte 0x08
		.byte 0x09
		.byte 0x00
		.byte 0x00	"); 
asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm ("		.byte 0x00
		.byte 0x0a
		.byte 0x0b
		.byte 0x0c	"); 
asm ("		.byte 0x0d
		.byte 0x0e
		.byte 0x0f
		.byte 0x00	"); 
asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");



/* dp		(P --- dictionary_free-point_pointer ) */
  
asm	("
	1:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \"dp\"
	.align 4, 0
	.int  0b
	.equ	dpCFA, .
	");

asm	(" .int addressCFA ");

dpaddy:		/* we have to poke a number in here at init time */
asm	(" .int 0	");



/* epa		(P --- &epa )  End Parse Area		*/
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 0
	.ascii \"epa\"
	.align 4, 0
	.int  1b
	.equ	epaCFA, .
	");

 /*  can't just keep these on a stack because of doHNC	  */
asm	(" .int addressCFA ");

epal:
asm	(" .int 0	");



/* hexdigits	(P  ---	 &array )  ascii val is/not a hex digit	  */
  
asm	("
	1:
	.byte 9
	.byte 0
	.byte 0
	.byte 0
	.ascii \"hexdigits\"
	.align 4, 0
	.int  0b
	.equ	hexdigitsCFA, .
	");

       /* could be dual-aligned for unicode. someday. word-size dependant? */
asm	(" .int addressCFA ");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm ("		.byte 0xff
		.byte 0xff
		.byte 0xff
		.byte 0xff	"); 
asm ("		.byte 0xff
		.byte 0xff
		.byte 0xff
		.byte 0xff	"); 
asm ("		.byte 0xff
		.byte 0xff
		.byte 0x00
		.byte 0x00	"); 
asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm ("		.byte 0x00
		.byte 0xff
		.byte 0xff
		.byte 0xff	"); 
asm ("		.byte 0xff
		.byte 0xff
		.byte 0xff
		.byte 0x00	"); 
asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");

asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");
 asm	(" .int 0 	");


/* latest		(P --- &HNC )  &HNC of most recent definition  */
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"latest\"
	.align 4, 0
	.int  1b
	.equ	latestCFA, .
	");

asm	(" .int addressCFA ");
		/* data word, does it's own Return.  */
latest:			/* C needs this address for init code	 */
asm	(" .int 0	");


/* pvar		(P --- &pvar)	scratch pointer for thread resolver  */
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 1
	.ascii \"pvar\"
	.align 4, 0
	.int  0b
	.equ	pvarCFA, .
	");

asm	(" .int addressCFA ");

 asm	(" .int 0	");


/* source		(P --- &input_FD) FD of source of interpreter input */
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"source\"
	.align 4, 0
	.int  1b
	.equ	sourceCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 256 ");


/* thispointer	(P --- &this ) current address in buffered parse area	*/
  
asm	("
	1:
	.byte 11
	.byte 0
	.byte 0
	.byte 0
	.ascii \"thispointer\"
	.align 4, 0
	.int  0b
	.equ	thispointerCFA, .
	");

asm	(" .int addressCFA ");

thisl:
asm	(" .int 0	");



/* toklen		token length	*/
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"toklen\"
	.align 4, 0
	.int  1b
	.equ	toklenCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");



/* targetstack	space for 8 addresses to resolve branches to */
  
asm	("
	1:
	.byte 11
	.byte 0
	.byte 0
	.byte 1
	.ascii \"targetstack\"
	.align 4, 0
	.int  0b
	.equ	targetstackCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 32 ");
	/* allot for a maximal pyte */


/* tb		(P --- &token ) where we stash an input token		*/
  
asm	("
	0:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \"tb\"
	.align 4, 0
	.int  1b
	.equ	tbCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");

asm	("	.org . + 1024 ");


/* tsi		counter for branches to resolve */
  
asm	("
	1:
	.byte 3
	.byte 0
	.byte 0
	.byte 1
	.ascii \"tsi\"
	.align 4, 0
	.int  0b
	.equ	tsiCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 0	");



/* USize		(P --- &USize ) offstack Size store for interpreter
					fetch as a pointer	  */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 1
	.ascii \"USize\"
	.align 4, 0
	.int  1b
	.equ	USizeCFA, .
	");

asm	(" .int addressCFA ");

asm	(" .int 4	");
				/* x86 cell size */


/* */
/*	THREADS...
*/
 /* TTTTTTTTTTTTTTTTTTTTTTTT============				*/

/* &		( a --- )  !dp	 append a into dictionary at &dp at Size  */
  
asm	("
	1:
	.byte 1
	.byte 0
	.byte 0
	.byte 0
	.ascii \"&\"
	.align 4, 0
	.int  0b
	.equ	appendCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/*  (P &dp)		( a )	*/
asm	(" .int pdupCFA ");
		/*  (P &dp &dp)		( a )	*/
asm	(" .int pfetchCFA ");
		/*  (P &dp dp)		( a )	*/
asm	(" .int storeCFA ");
		/*  (P &dp dp)		()	*/
asm	(" .int pplussCFA ");
		/*  (P &dp dp')		()	*/
asm	(" .int pswapCFA ");
		/*  (P dp' &dp)			*/
asm	(" .int pstoreCFA ");
		/*  (P dp' &dp)		!dp	*/
asm	(" .int twopdropCFA ");
		/*  (P;)			*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* &literal	append literal  ( # --- )   !dp  */

  
asm	("
	0:
	.byte 8
	.byte 0
	.byte 0
	.byte 0
	.ascii \"&literal\"
	.align 4, 0
	.int  1b
	.equ	appendlitCFA, .
	");


asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int literalCFA ");

asm	(" .int TOpCFA ");

asm	(" .int subCFA	");

 asm	(" .int appendpCFA ");

asm	(" .int getSizeCFA ");
		/* ( # )	(P Size )		*/
asm	(" .int alignCFA ");
		/* ( # )	(P alignedSize)		*/
asm	(" .int subCFA	");

 asm	(" .int appendpCFA ");
		/* ( # )	(P )		!dp	*/
asm	(" .int subCFA	");

 asm	(" .int appendCFA ");
		/* ( )		(P )		!dp	*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* &word		(P &HNC ||| )  append a word's CFA.   */
  
asm	("
	1:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"&word\"
	.align 4, 0
	.int  0b
	.equ	appendwordCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int ifatomCFA ");
		/* (P &HNC )	( flag)		*/
asm	(" .int noCFA ");

	asm	(" .int	 .  + 7 *4  ");

			/*		  is an atom	*/
asm	(" .int pdupCFA ");
		/* (P &HNC &HNC)		*/
asm	(" .int TOcodeCFA ");
		/* (P &HNC CFA)			*/
asm	(" .int subCFA	");
			/* (P &HNC CFA)			*/
 asm	(" .int appendpCFA ");
		/* (P &HNC CFA)			*/
asm	(" .int pdropCFA ");
		/* (P &HNC )			*/
asm	(" .int ReturnCFA ");


			/* (P &HNC )	is a thread	*/
	asm	(" .int literalCFA ");

	  asm	(" .int 4	");
		/* sub's CFA as literal. Hmmmmm.	*/
	  asm	(" .int subCFA ");
		/* (P &HNC )	( subCFA )		*/
	asm	(" .int TOpCFA ");
			/* (P &HNC subCFA )			*/
	asm	(" .int subCFA	");

	 asm	(" .int appendpCFA ");
		/* (P &HNC subCFA )			*/
	asm	(" .int pdropCFA ");
		/* (P &HNC )			*/
	asm	(" .int pdupCFA ");
		/* (P &HNC &HNC )		*/
	asm	(" .int TOcodeCFA ");
		/* (P &HNC CFA )		*/
	asm	(" .int subCFA	");

	 asm	(" .int appendpCFA ");

	asm	(" .int pdropCFA ");
		/* (P &HNC )			*/
	asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* &pliteral	thread-time -> "&pliteral lit# " THREADING	 */
  
asm	("
	0:
	.byte 9
	.byte 1
	.byte 0
	.byte 0
	.ascii \"&pliteral\"
	.align 4, 0
	.int  1b
	.equ	appendpliteralCFA, .
	");


asm	(" .int sTOrCFA ");

asm	(" .int cells_bdCFA ");

asm	(" .int literalCFA ");

  asm	(" .int 4	");
		/* pliteral's CFA as literal. Hmmmmm.	*/
  asm	(" .int pliteralCFA ");
		/* (  plitCFA )				*/
asm	(" .int TOpCFA ");
			/* (P plitCFA )				*/
asm	(" .int subCFA	");

 asm	(" .int appendpCFA ");
		 /* (P plitCFA )				*/
asm	(" .int pdropCFA ");
		/*					*/
asm	(" .int subCFA	");

  asm	(" .int tokenCFA ");

asm	(" .int subCFA	");

  asm	(" .int hexnumberCFA ");

asm	(" .int subCFA	");

  asm	(" .int appendCFA ");

asm	(" .int rTOsCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* &p		(P bla |||  )	append pointer bla to dictionary  !dp	*/
  
asm	("
	1:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \"&p\"
	.align 4, 0
	.int  0b
	.equ	appendpCFA, .
	");


asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/*	(P bla &dp)			*/
asm	(" .int pdupCFA ");
		/*	(P bla &dp &dp)			*/
asm	(" .int pTOrCFA ");
		/*	(P bla &dp)	(R; &dp)	*/
asm	(" .int pfetchCFA ");
		/*	(P bla	dp)	(R; &dp)	*/
asm	(" .int pstoreCFA ");
		/*	(P bla	dp)	(R; &dp)   !HERE*/
asm	(" .int ppluscCFA ");
		/*	(P bla	dp')	(R; &dp)	*/
asm	(" .int rTOpCFA ");
		/*	(P bla	dp' &dp)		*/
asm	(" .int pstoreCFA ");
		/*	(P bla	dp' &dp)    !dp		*/
asm	(" .int twopdropCFA ");

asm	(" .int ReturnCFA ");




 /* TTTTTTTTTTTTTTTTTTTTTTTT============		*/

/* abs		( A --- |A| )			*/
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 0
	.ascii \"abs\"
	.align 4, 0
	.int  1b
	.equ	absCFA, .
	");

asm	(" .int dupCFA ");
			/* (A A)		*/
 asm	(" .int ifpositiveCFA ");
	       /* (A flag)	       */
asm	(" .int yesCFA ");

 asm	(" .int	 .  + 2 *4  ");

asm	(" .int negateCFA ");

asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttttt============ */

/* allot		( o/s --- )  move dp up o/s bytes. iffy at > 256*/
  
asm	("
	1:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"allot\"
	.align 4, 0
	.int  0b
	.equ	allotCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int dpCFA ");

asm	(" .int pdupCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int pluspCFA ");

asm	(" .int pswapCFA ");

asm	(" .int pstoreCFA ");

asm	(" .int ReturnCFA ");






 /* tttttttttttttttttttttt============ */

/* ascii->digit	( --- nybble )	(P char	 ||| )	  */
  
asm	("
	0:
	.byte 12
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ascii->digit\"
	.align 4, 0
	.int  1b
	.equ	asciiTOdigitCFA, .
	");

asm	(" .int fetchCFA ");
		/*   ( pyte		(P		*/
asm	(" .int maskbyteCFA ");
		/*   ( byte		(P		*/
asm	(" .int subCFA	");

 asm	(" .int digitsCFA ");
		/*   ( byte		(P  array	*/
asm	(" .int pluspCFA ");
		/*   (			(P indexed	*/
asm	(" .int fetchCFA ");
		/*   ( nybblepyte	(P		*/
asm	(" .int pdropCFA ");

asm	(" .int maskbyteCFA ");
		/*   ( nybble		(P		*/
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttttttttttttttttttttttttttttttt============ */
/* beep		emit a 7, ASCII bel, C \a		*/
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"beep\"
	.align 4, 0
	.int  0b
	.equ	beepCFA, .
	");

asm	(" .int literalCFA ");
 asm	(" .int 4	");
 asm	(" .int 7	");
 asm	(" .int emitCFA ");
 asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* bench		bogoloop, 100,000 empty loop iterations.  */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"bench\"
	.align 4, 0
	.int  1b
	.equ	benchCFA, .
	");
	/*  (P &dp &dp)		( a )	*/

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 100000	");

asm	(" .int TOrCFA ");


asm	("	.equ bogo_one, .	");

	asm	(" .int minusaskrCFA ");

     asm	(" .int noCFA ");

	 asm	(" .int	 .  + 3 *4  ");

asm	(" .int ellCFA ");

 asm	("	.int bogo_one	");


asm	(" .int rTOpCFA ");

asm	(" .int pdropCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */
/* blank		emit a blank character, 0x20, a space  */
  
asm	("
	1:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"blank\"
	.align 4, 0
	.int  0b
	.equ	blankCFA, .
	");

asm	(" .int literalCFA ");
 asm	(" .int 4	");
 asm	(" .int 32	");
 asm	(" .int emitCFA ");
 asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttt============ */

/* :		": ccc H3sm.. ; " name/link ccc, set target 0, start threader */
  
asm	("
	0:
	.byte 1
	.byte 0
	.byte 0
	.byte 0
	.ascii \":\"
	.align 4, 0
	.int  1b
	.equ	colonCFA, .
	");

asm	(" .int subCFA	");
			/* wipe branch-resolver stacks		*/
  asm	(" .int initresolveCFA ");

asm	(" .int subCFA	");

 asm	(" .int headerCFA ");

asm	(" .int subCFA	");

 asm	(" .int linkCFA ");

asm	(" .int subCFA	");
			/* stash newguy's CFA as target0	*/
  asm	(" .int targetCFA ");

asm	(" .int dropCFA ");

asm	(" .int subCFA	");

 asm	(" .int threadloopCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/
/* cr		emit a (unix) newline				*/
  
asm	("
	1:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \"cr\"
	.align 4, 0
	.int  0b
	.equ	crCFA, .
	");

asm	(" .int literalCFA ");
 asm	(" .int 4	");
 asm	(" .int 10	");
 asm	(" .int emitCFA ");
 asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* dofile		STUB					*/
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 1
	.ascii \"dofile\"
	.align 4, 0
	.int  1b
	.equ	dofileCFA, .
	");

asm	(" .int ReturnCFA ");



   /* ttttttttttttttttttttttttttttttttttttttt============ */
/* dump		(P start --- end ) hexdump 256 bytes, continuable  */
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"dump\"
	.align 4, 0
	.int  0b
	.equ	dumpCFA, .
	");


asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 16	");

asm	(" .int TOrCFA ");

asm	("	.equ dump_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int dumplineCFA ");
		/*  (P	a'' )		*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 124	");

	asm	(" .int emitCFA ");

	asm	(" .int minusaskrCFA ");

	asm	(" .int noCFA ");

		asm	(" .int	 .  + 3 *4  ");

asm	(" .int ellCFA ");

   asm	("	.int dump_one	");

		asm	(" .int rdropCFA ");

		asm	(" .int subCFA	");

		  asm	(" .int crCFA ");

		asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttttttttttttttt=========== */

/* dumpline	(P a --- a+16 )	 display 16 bytes from a    */
  
asm	("
	0:
	.byte 8
	.byte 0
	.byte 0
	.byte 1
	.ascii \"dumpline\"
	.align 4, 0
	.int  1b
	.equ	dumplineCFA, .
	");


asm	(" .int sTOrCFA ");

asm	(" .int cellsCFA ");

asm	(" .int pdupCFA ");

asm	(" .int subCFA	");

 asm	(" .int crCFA ");

asm	(" .int subCFA	");

 asm	(" .int printpCFA ");

asm	(" .int subCFA	");
	asm	(" .int blankCFA ");

asm	(" .int subCFA	");
	asm	(" .int blankCFA ");

asm	(" .int bytesCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");
 asm	(" .int fetchplusCFA ");
	asm	(" .int printhexCFA ");

			/* (P  a  a+16)		*/
asm	(" .int subCFA	");
	asm	(" .int blankCFA ");
 asm	(" .int subCFA	");
	asm	(" .int blankCFA ");


asm	(" .int literalCFA ");

  asm	(" .int 4	");

  asm	(" .int 124	");

asm	(" .int emitCFA ");

asm	(" .int pswapCFA ");
				/* (P a+16 a )		*/

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 15	");

asm	(" .int subCFA	");

 asm	(" .int printcharsCFA ");

asm	(" .int rTOsCFA ");

asm	(" .int pdropCFA ");
				/* (P a+16)		*/
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* dpdump		(P --- dp+0xc0 )   dump from dp - 0x80	*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"dpdump\"
	.align 4, 0
	.int  0b
	.equ	dpdumpCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 0x80	");

asm	(" .int negateCFA ");

asm	(" .int pluspCFA ");

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 7	");

asm	(" .int TOrCFA ");


asm	("	.equ mdump_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int dumplineCFA ");
		/*  (P	a'' )		*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 124	");

	asm	(" .int emitCFA ");

	asm	(" .int minusaskrCFA ");

	asm	(" .int noCFA ");

    asm	(" .int	 .  + 3 *4  ");

asm	(" .int ellCFA ");

   asm	("	.int mdump_one	");

asm	(" .int rdropCFA ");

asm	(" .int subCFA	");

 asm	(" .int crCFA ");

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 8	");

asm	(" .int TOrCFA ");


asm	("	.equ mdump2_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int dumplineCFA ");
		/*  (P	a'' )		*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 124	");

	asm	(" .int emitCFA ");

	asm	(" .int minusaskrCFA ");

	asm	(" .int noCFA ");

    asm	(" .int	 .  + 3 *4  ");

asm	(" .int ellCFA ");

   asm	("	.int mdump2_one	");

asm	(" .int rdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int crCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* delimit	(P dest |||) ( delimiter --- count did ) stash input */
  
asm	("
	0:
	.byte 7
	.byte 0
	.byte 0
	.byte 0
	.ascii \"delimit\"
	.align 4, 0
	.int  1b
	.equ	delimitCFA, .
	");

asm	(" .int pdupCFA ");
			/* (P dest0 dest')  ( delim )		*/
asm	("	.equ while_this_isnt_delim_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int thisCFA ");
		/* (P dest0 dest')  (delim this)	*/
	asm	(" .int subCFA	");

	  asm	(" .int stepCFA ");
		/* (P dest0 dest')  (delim this EOI)	*/
		asm	(" .int yesCFA ");

		asm	(" .int	 .  + 11 *4  ");

	asm	(" .int overCFA ");

	asm	(" .int overCFA ");
		/* (P dest0 dest')  (delim this del ths)*/
	asm	(" .int maskbyteCFA ");
			
 	asm	(" .int ifequalCFA ");
		/* (P dest0 dest')  (delim this =)	*/	
			asm	(" .int yesCFA ");
 	/* (P dest0 dest') (delim this )*/
			asm	(" .int	 .  + 12 *4  ");

	asm	(" .int storeCFA ");
		/* (P dest0 dest')  (delim )		*/
	asm	(" .int pplusbCFA ");
		/* (P dest0 dest++) (delim )		*/
asm	(" .int ellCFA ");

asm	("	.int while_this_isnt_delim_one	");

		asm	(" .int dropCFA ");

		asm	(" .int dropCFA ");

		asm	(" .int gapCFA ");

		asm	(" .int negateCFA ");

		asm	(" .int pdropCFA ");

		asm	(" .int falseCFA ");

		asm	(" .int ReturnCFA ");

			asm	(" .int dropCFA ");

			asm	(" .int dropCFA ");

			asm	(" .int gapCFA ");

			asm	(" .int negateCFA ");

			asm	(" .int pdropCFA ");

			asm	(" .int trueCFA ");

			asm	(" .int ReturnCFA ");

			

 /* ttttttttttttttttttttttt============ */

/* drowod		(P HNC --- ???) (???)  (R ?????) !USize	 Interpreter only*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 1
	.ascii \"drowod\"
	.align 4, 0
	.int  0b
	.equ	drowodCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int USizeCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int bumpCFA ");

asm	(" .int pTOsCFA ");

asm	(" .int dropCFA ");

asm	(" .int doHNCCFA ");
		/*U (P ???) (???) (R???)			*/
 /* we may have changed Size, so we save it for the next time we reenter
user context. */

asm	(" .int sTOpCFA ");
		/*  (P ??? USize) (???) (R???)			*/
asm	(" .int subCFA	");

 asm	(" .int USizeCFA ");
		/*  (P ??? USize &USize) (???) (R???)		*/
asm	(" .int pstoreCFA ");
		/*  (P ??? USize &USize) (???) (R???)		*/
asm	(" .int twopdropCFA ");
		/*  (P ???) (???) (R???)			*/
asm	(" .int cells_bdCFA ");
		   /*	4 !Size				   */
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt ============			*/

/* finish		( --- EOI) !toklen !tb !this	get rest of token */
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"finish\"
	.align 4, 0
	.int  1b
	.equ	finishCFA, .
	");

				/* we come in on a nonwhite from snug	*/
asm	(" .int subCFA	");

  asm	(" .int tbCFA ");
			/* (P tokenbuffer)		*/
asm	(" .int subCFA	");

  asm	(" .int thisCFA ");
			/*  (P tb)	( this )		*/
asm	(" .int storeCFA ");
			/*  (P tb) <-- ()			*/
asm	(" .int pplusbCFA ");
			/*  (P tokenbuf_next) ()		*/
asm	(" .int oneCFA ");
				/*  (P tokenbuf_next) (initial_toklen=1)*/
asm	("	.equ while_nonwhite_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int stepCFA ");
		/* this++ (P tb) (toklen EOI)		*/
	asm	(" .int noCFA ");
			/* (P tb)   ( toklen )			*/
	asm	(" .int	 .  + 5 *4  ");

		asm	(" .int dropCFA ");
	/* ran out of input			*/
		asm	(" .int pdropCFA ");

		asm	(" .int trueCFA ");
	/* ( EOI=true )				*/
		asm	(" .int ReturnCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int thispointerCFA ");
	/*  (P tb' &this ) ( toklen )		*/
	asm	(" .int pfetchCFA ");
		/*  (P tb' this )  ( toklen )		*/
	asm	(" .int subCFA	");

	  asm	(" .int ifcharCFA ");
		/*  (P tb' )	(toklen	 ischar)
				ifchar does a pdrop. Doh! BAD!		*/
		asm	(" .int noCFA ");
		/*  (P tb' ) ( toklen )		*/
		asm	(" .int	 .  + 9 *4  ");

	asm	(" .int oneCFA ");
			/* (P tb)   ( toklen 1 )		*/
	asm	(" .int plusCFA ");
		/* (P tb)   ( toklen'  )		*/
	asm	(" .int subCFA	");

	  asm	(" .int thisCFA ");
		/*  (P tb )  (toklen' this )		*/
	asm	(" .int storeCFA ");
		/*  (P !tb )  <--  (toklen' )		*/
	asm	(" .int pplusbCFA ");
		/*  (P tokenbuf_next)	(toklen' )	*/
asm	(" .int ellCFA ");

asm	("	.int while_nonwhite_one	");

		asm	(" .int subCFA	");

		  asm	(" .int toklenCFA ");
	/* (P tb' toklen)     (toklen)	*/
		asm	(" .int storeCFA ");
	/* (P tb' !toklen) <--/		*/
		asm	(" .int twopdropCFA ");
	/* (P tb' !toklen)		*/
		asm	(" .int falseCFA ");
	/* ( EOI=false)			*/
		asm	(" .int ReturnCFA ");




 /* ttttttttttttttttttttttt============ */

/* <-code		fromcode (P CFA --- HNC )  ( --- found_flag )
				see needs this.			*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"<-code\"
	.align 4, 0
	.int  0b
	.equ	fromcodeCFA, .
	");


asm	(" .int pminuscCFA ");
		/* (P lfa_of_sought)				*/
asm	(" .int subCFA	");

  asm	(" .int latestCFA ");
		/* (P sought &HNC)				*/
asm	(" .int pfetchCFA ");
		/* (P sought HNC)
asm	(" .int pdupCFA ");

asm	(" .int pTOrCFA ");
		/* (P sought HNC) (R HNC)			*/
asm	(" .int TOlinkCFA ");
		/* (P sought lfa) (R HNC)			*/

asm	("	.equ fromcfa_one, .	");

	asm	(" .int rdropCFA ");
	/* (P sought lfa) (R)				*/
	asm	(" .int pfetchCFA ");
	/* (P sought HNCp) (R)				*/
	asm	(" .int fetchCFA ");

	asm	(" .int noCFA ");
		/* yes, name has length				*/
	asm	(" .int	 .  + 12 *4  ");
	/* AARDVARK!					*/
	asm	(" .int pdupCFA ");

	asm	(" .int pTOrCFA ");
	/* (P sought HNC) (R HNC)			*/
	asm	(" .int TOlinkCFA ");
	/* (P sought lfa) (R HNC)			*/
	asm	(" .int gapCFA ");
		/* (P sought lfa) (R HNC) ( same?)		*/
	asm	(" .int flagCFA ");

       asm	(" .int yesCFA ");
		/* (P sought lfa) (R HNC) () yup. Found it.	*/
	  asm	("	.int fromcfa_one	");
 /* yes, different. try again.			*/

asm	(" .int trueCFA ");
		/* found					*/
asm	(" .int twopdropCFA ");
		/* (P)	 (R HNC)				*/
asm	(" .int rTOpCFA ");
		/* (P HNC)	(R)				*/
asm	(" .int ReturnCFA ");


asm	(" .int falseCFA ");
		/* no word found with that "CFA"		*/
asm	(" .int twopdropCFA ");
		/* (P)	 (R HNC)				*/
asm	(" .int rTOpCFA ");
		/* (P HNC)	(R)				*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */
/* H3m		no-op thread, simplifies commandline handling  */
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 0
	.ascii \"H3m\"
	.align 4, 0
	.int  1b
	.equ	H3mCFA, .
	");

asm	(" .int ReturnCFA ");



 /* tttttmttttttttttttttttttttttttttttttttttt============ */

/* hexnumber	( --- # )  from current token			*/
  
asm	("
	1:
	.byte 9
	.byte 0
	.byte 0
	.byte 0
	.ascii \"hexnumber\"
	.align 4, 0
	.int  0b
	.equ	hexnumberCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int toklenCFA ");
			/*			(P &toklen)	*/
asm	(" .int fetchCFA ");
			/* ( len )		(P )		*/
asm	(" .int pdropCFA ");

asm	(" .int subCFA	");

 asm	(" .int tbCFA ");
				/* ( len )		(P tb )		*/
asm	(" .int zeroCFA ");
			/* ( len  accum0)	(P bt)		*/
asm	("	.equ while_hexchar_one, .	");

	asm	(" .int overCFA ");
		/* ( len accum len)	(P char)	*/
	asm	(" .int noCFA ");
			/* ( len accum)		(P char)	*/
		asm	(" .int	 .  + 15 *4  ");

	asm	(" .int literalCFA ");
		/* ( len  accum)	(P char)	*/
	asm	(" .int 4	");
			/* ( len  accum)	(P char)	*/
	asm	(" .int 4	");
			/* ( len  accum	 4)	(P char)	*/
	asm	(" .int upshiftCFA ");
		/* ( len  accum<4)	(P char)	*/
	asm	(" .int subCFA	");

	  asm	(" .int asciiTOdigitCFA ");
	/* ( len  accum< dig)	(P char)	*/
	asm	(" .int ORCFA ");
			/* ( len  accum<+)	(P char)	*/
	asm	(" .int swapCFA ");
		/* ( accum  cnt)	(P char)	*/
	asm	(" .int onbitsCFA ");
		/* ( accum  cnt -1)	(P  char)	*/
	asm	(" .int plusCFA ");
		/* ( accum  cntd)	(P  char)	*/
	asm	(" .int swapCFA ");
		/* ( cntdn  accum)	(P  char)	*/
	asm	(" .int pplusbCFA ");
		/* ( cntdn  accum)	(P char')	*/
asm	(" .int ellCFA ");

  asm	("	.int while_hexchar_one	");

				/* ( len accum)		(P char)	*/
		asm	(" .int pdropCFA ");
	/* ( len accum)		(P )		*/
		asm	(" .int swapCFA ");
	/* ( accu  len)		(P )		*/
		asm	(" .int dropCFA ");
	/* ( accu)		(P )		*/
		   asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* ifchar		(p; str --- )  ( --- flag )  true if byte > 32	*/
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifchar\"
	.align 4, 0
	.int  1b
	.equ	ifcharCFA, .
	");

asm	(" .int fetchCFA ");
		/* (P str )   ( char )	(R: ParentSize ) !Size */
asm	(" .int pdropCFA ");

asm	(" .int maskbyteCFA ");

asm	(" .int literalCFA ");

  asm	(" .int 4	");

  asm	(" .int 32	");
		/* ( byte  32  )		*/
asm	(" .int negateCFA ");
		/* ( byte  -32)			*/
asm	(" .int plusCFA ");
		/* ( byte-32)			*/
asm	(" .int zeroCFA ");
		/* ( byte-32 0 )		*/
asm	(" .int maxCFA ");
			/* ( 0_or_ischar)		*/
asm	(" .int flagCFA ");
		/* ( flag ) true if non-white	*/
asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttttttt  ============ */

/* ifword		(P ---	HNC|null) ( --- flag )	*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifword\"
	.align 4, 0
	.int  0b
	.equ	ifwordCFA, .
	");

asm	(" .int subCFA	");

   asm	(" .int tbCFA ");
		/*   (P tb )				*/
asm	(" .int subCFA	");

   asm	(" .int toklenCFA ");
		/*   (P bt toklen)		(R oS
				this limits toklen to 255.		*/
asm	(" .int fetchCFA ");
		/*   (P bt )  ( toklen)		(R oS		*/
asm	(" .int pdropCFA ");

asm	(" .int maskbyteCFA ");

asm	(" .int subCFA	");

   asm	(" .int latestCFA ");

asm	(" .int pfetchCFA ");
		/*   (P bt latest ) ( toklen )	(R oS		*/
asm	("	.equ while_dictionary_one, .	");

	asm	(" .int fetchCFA ");
		/*   (P bt latest ) ( len  namebyte )	*/
	asm	(" .int maskbyteCFA ");

	asm	(" .int yesCFA ");
			/*   (P bt latest ) ( len)		*/
	asm	(" .int	 .  + 5 *4  ");

		asm	(" .int dropCFA ");
	/*   (P bt latest )	()		*/
		asm	(" .int falseCFA ");
	/*   (P bt latest )	( false )	*/
		asm	(" .int twopdropCFA ");
	/*   (P) = NULL		(false )	*/
		asm	(" .int ReturnCFA ");

	asm	(" .int fetchCFA ");
		/*   (P tbt  last) ( toklen  namelen  ) */
	asm	(" .int maskbyteCFA ");
	
	asm	(" .int overCFA ");
		/*   (P tb  last ) ( toklen  namelen toklen*/
	asm	(" .int ifequalCFA ");
		/*   (P tb  last ) ( toklen  flag )	*/
	asm	(" .int noCFA ");
			/*   (P tb  last ) ( toklen )		*/
		asm	(" .int	 .  + 13 *4  ");
	/* not the same length as this name	*/
					/* same length	*/
	asm	(" .int ppluscCFA ");
		/*   (P begintoken  name ) ( toklen )	*/
	asm	(" .int subCFA	");

	  asm	(" .int ifbytesegmentsequalCFA ");
	/*   (P tb name ) ( toklen  flag )*/
	asm	(" .int noCFA ");

			asm	(" .int	 .  + 7 *4  ");
	/* not this word		*/
	
	asm	(" .int pswapCFA ");
		/*   (P name tb)	( toklen )	*/
	asm	(" .int pdropCFA ");
		/*   (P name)		( toklen )	*/	
	asm	(" .int pminuscCFA ");
		/*   (P HNC )		( toklen )	*/
	asm	(" .int dropCFA ");
		/*   (P HNC ) ()			*/
	asm	(" .int trueCFA ");
		/*   (P HNC ) ( TRUE )			*/
	asm	(" .int ReturnCFA ");

						/*from asksegmentsequal */
			asm	(" .int pminuscCFA ");
	/*   (P begintok this_HNC)  (toklen )*/ 
					/*   from askequal		*/
		asm	(" .int TOlinkCFA ");
		/*   "previous"			*/
		asm	(" .int pfetchCFA ");
		/*   (P bt HNC' ) (toklen )	*/
asm	(" .int ellCFA ");

asm	("	.int while_dictionary_one	");



 /* tttttttttttttttttttttttt============ */

/* ifabbreviation	(P --- HNC|NULL)  ( --- flag ) "'abbreviation toke. "*/
  
asm	("
	0:
	.byte 14
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifabbreviation\"
	.align 4, 0
	.int  1b
	.equ	ifabbreviationCFA, .
	");


asm	(" .int subCFA	");

   asm	(" .int tbCFA ");
		/*   (P &bt)					*/
asm	(" .int sTOrCFA ");
		/*   (P bt )			(R Size		*/
asm	(" .int bytes_bdCFA ");
		/* Size = 1	!!!bytes			*/
asm	(" .int subCFA	");

   asm	(" .int toklenCFA ");
		/*   (P bt toklen)			(R oS	*/
asm	(" .int fetchCFA ");
		/*   (P bt toklen)	( len)		(R oS	*/
asm	(" .int pdropCFA ");
		/*   (P bt )		( len)		(R oS	*/
			/* is it period-terminated?			*/
asm	(" .int pdupCFA ");
		/*   (P bt bt)	( len)		(R oS)		*/
asm	(" .int oneCFA ");

asm	(" .int subCFA	");

 asm	(" .int minusCFA ");

asm	(" .int dupCFA ");
			/*   (P bt )	( len-1 len)	(R oS)		*/
asm	(" .int pluspCFA ");
		/*   (P bt et)	( len-1)	(R oS)		*/
asm	(" .int fetchCFA ");
		/*   (P bt et)	( len lastchar) (R oS)		*/
asm	(" .int literalCFA ");
		/*   (P bt et)	( len lastchar) (R oS)		*/
  asm	(" .int 4	");

  asm	(" .int 46	");
		/*   (P bt et)	( len lastchar .) (R oS)	*/
asm	(" .int ifequalCFA ");
	       /*   (P bt et)  ( len flag)	(R oS)	       */
asm	(" .int yesCFA ");
			/*   (P bt et)	( len)		 (R oS)		*/
 asm	(" .int	 .  + 6 *4  ");

asm	(" .int dropCFA ");
	       /* NOT AN ABBREVIATION, exuent			*/
asm	(" .int falseCFA ");
		/*   (P bt et)	(false)		(R oS)		*/
asm	(" .int twopdropCFA ");

asm	(" .int rTOsCFA ");

asm	(" .int ReturnCFA ");
		/* (false)					*/
		       /* TOKEN IS AN ABBREVIATION	(R oS)		*/
asm	(" .int pdropCFA ");
		/*   (P bt )	(len-1)		(R oS)		*/
asm	(" .int subCFA	");

   asm	(" .int latestCFA ");

asm	(" .int pfetchCFA ");
		/*   (P bt latest ) (len )  (R oS)		*/
			/* seed ds ?hit and rs ?HNC			*/
asm	(" .int falseCFA ");

asm	(" .int swapCFA ");
		/*   (P bt latest ) (?hit  len) (R oS)		*/
asm	(" .int pdupCFA ");

asm	(" .int pTOrCFA ");
		/*   (P bt latest ) (?hit  len) (R ?HNC oS)	*/

asm	("	.equ ifabbr_one, .	");


asm	(" .int fetchCFA ");
		/*   (P bt latest ) (hit? toklen  namebyte )	*/
asm	(" .int yesCFA ");
			/* is namelen byte <>0?				*/
   asm	(" .int	 .  + 12 *4  ");
		/* yes. more dictionary to do.			*/
			/* no. aardvark. Finish up.			*/
asm	(" .int twopdropCFA ");
		/*   (P )		(hit? toklen)		*/
asm	(" .int dropCFA ");
		/*   (P )		(hit? )			*/
asm	(" .int dupCFA ");
			/*   (P )		(hit? hit?)		*/
asm	(" .int noCFA ");
			/*   (P )		(hit?)			*/
 asm	(" .int	 .  + 4 *4  ");

			/* was a word					*/
asm	(" .int rTOpCFA ");
		/*   (P &HNC)		(true)			*/
asm	(" .int rTOsCFA ");
		/*	(P &HNC)	(true)			*/
asm	(" .int ReturnCFA ");

			/* wasn't a word				*/
asm	(" .int rdropCFA ");
		/*   (P) = NULL		(false)			*/
asm	(" .int rTOsCFA ");

asm	(" .int ReturnCFA ");

			/*   (P bt  last ) (hit? len )			*/
			/* substring comparison, hop over to name	*/
asm	(" .int ppluscCFA ");
		/*   (P bt name ) (hit? len )			*/
asm	(" .int subCFA	");
		/*  {{{(P beg.	beg. ||| --- ) ( len ||| --- flag}}}	*/
   asm	(" .int ifsegmentsequalCFA ");
 /*   (P bt  name ) (hit? len  flag )	       */
	asm	(" .int noCFA ");
		/* sub-match?					*/
	asm	(" .int	 .  + 13 *4  ");
	/* nope.					*/
			/* yup. this is one possible completion		*/
asm	(" .int pminuscCFA ");
		/*   (P bt  &HNC_hit )	(hit? len)		*/
asm	(" .int rdropCFA ");
		/* overwrite rs?HNC				*/
asm	(" .int pdupCFA ");

asm	(" .int pTOrCFA ");
		/*    "	     "					*/
			/* overwrite ?hit with true			*/
asm	(" .int swapCFA ");
		/* (P bt &HNC_hit )	(len ?hit)		*/
asm	(" .int dropCFA ");
		/* (P bt &HNC_hit )	(len)			*/
asm	(" .int trueCFA ");
		/* (P bt &HNC_hit )	(len true)		*/
asm	(" .int swapCFA ");
		/* (P bt &HNC_hit )	(true len)		*/
asm	(" .int TOlinkCFA ");
		/* (P bt LFA  )		(true len)		*/
asm	(" .int pfetchCFA ");
		/* (P bt HNC' )		(true len)		*/
asm	(" .int ellCFA ");

   asm	("	.int ifabbr_one	");

	asm	(" .int pminuscCFA ");

	asm	(" .int TOlinkCFA ");
		/* (P bt LFA  ) (?hit len )		*/
	asm	(" .int pfetchCFA ");
		/* (P bt HNC' ) (?hit len )		*/
	asm	(" .int ellCFA ");

	   asm	("	.int ifabbr_one	");



 /* ttttttttttttttttttttttttttttttttttttttttttttt============ */

/* ifatom		(P &HNC |||)   ( --- flag ) true if HNC is an atom */
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifatom\"
	.align 4, 0
	.int  0b
	.equ	ifatomCFA, .
	");

asm	(" .int pdupCFA ");
		/* (P &HNC ||| &HNC			*/
asm	(" .int pfetchCFA ");
		/* (P &HNC ||| HNC			*/
asm	(" .int pTOCFA ");
			/* (P HNC |||) ( HNC )			*/
asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 0x00800000	");
	/* (P HNC |||) ( HNC abit)		*/
asm	(" .int ANDCFA ");
			/* (P HNC |||) ( ?atom )		*/
asm	(" .int flagCFA ");
		/* (P HNC |||) ( flag )			*/
asm	(" .int ReturnCFA ");


 /* ttttttttttttttttttttttt============ */

/* ifhexnumber	( --- flag )  @'s toklen and tb */
  
asm	("
	0:
	.byte 11
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifhexnumber\"
	.align 4, 0
	.int  1b
	.equ	ifhexnumberCFA, .
	");

asm	(" .int subCFA	");

   asm	(" .int tbCFA ");

asm	(" .int subCFA	");

   asm	(" .int toklenCFA ");

asm	(" .int fetchCFA ");
			/*  (toklen )	(P bt )		*/
asm	(" .int pdropCFA ");

asm	(" .int oneCFA ");
				/*  (toklen 1)	(P bt )		*/
asm	(" .int plusCFA ");
			/*  (pre-count) (P bt )		*/
asm	("	.equ ifhexnum_one, .	");

	asm	(" .int onbitsCFA ");
		/*   ( count -1 )   (P bt )	*/
	asm	(" .int plusCFA ");
		/*   ( count')		(P bt ) */
	asm	(" .int dupCFA ");
			/*   ( count' count')	(P bt ) */
	asm	(" .int noCFA ");
			/*   ( count')		(P bt ) */
		asm	(" .int	 .  + 8 *4  ");

	asm	(" .int subCFA	");

	asm	(" .int ifhexCFA ");
		/*   ( count' flag )	(P bt )		*/
	asm	(" .int pplusbCFA ");
		/*   ( count' flag )	(P cur.char )	*/
	asm	(" .int noCFA ");
			/*   ( count' )		(P cur.char )	*/
	asm	(" .int	 .  + 7 *4  ");

asm	(" .int ellCFA ");
				/* is hex digit. ( count') (P cur.char )*/
asm	("	.int ifhexnum_one	");

		asm	(" .int dropCFA ");
	/* done, is number   () (Pbeg.tok)	*/
		asm	(" .int pdropCFA ");
	/*   ()		(P)			*/
		asm	(" .int trueCFA ");
	/*   (xxxff)	(P)			*/
		asm	(" .int ReturnCFA ");

	asm	(" .int dropCFA ");
		/* not a number (count	(P array )	*/
	asm	(" .int pdropCFA ");
		/*   (	(P)				*/
	asm	(" .int falseCFA ");
		/*   (	xxx00 )		(P)		*/
	asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttttttttttttttttttttttt============ */

/* iffile		STUB ( --- flag ) "iffile  token "	*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 1
	.ascii \"iffile\"
	.align 4, 0
	.int  0b
	.equ	iffileCFA, .
	");

asm	(" .int falseCFA ");

asm	(" .int ReturnCFA ");


 /* ttttttttttttttttttttttt============ */

/* ifthreader	(P &HNC |||) ( --- flag ) true if threader word */
  
asm	("
	0:
	.byte 10
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifthreader\"
	.align 4, 0
	.int  1b
	.equ	ifthreaderCFA, .
	");

asm	(" .int pdupCFA ");
		/* (P &HNC ||| &HNC			*/
asm	(" .int pfetchCFA ");
		/* (P &HNC ||| HNC			*/
asm	(" .int pTOCFA ");
			/* (P HNC |||) ( HNC )			*/
asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 0x00000100	");
	/* (P HNC |||) ( HNC abit)		*/
asm	(" .int ANDCFA ");
			/* (P HNC |||) ( ?atom )		*/
asm	(" .int flagCFA ");
		/* (P HNC |||) ( flag )			*/
asm	(" .int ReturnCFA ");


 /* ttttttttttttttttttttttttttttttttttttttttttttttttttttt============ */

/* ifpending	(--- flag) true if not finished refill	  @this @epa	*/
  
asm	("
	1:
	.byte 9
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifpending\"
	.align 4, 0
	.int  0b
	.equ	ifpendingCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int thispointerCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int subCFA	");

  asm	(" .int epaCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int gapCFA ");

asm	(" .int flagCFA ");

asm	(" .int twopdropCFA ");

asm	(" .int ReturnCFA ");


 /* ttttttttttttttttttttttttttttttttttttttttttttttttttttt============ */

/* if$=		(P beg.	 beg. ||| --- ) ( len ||| --- flag ) are memory segments =? */
  
asm	("
	0:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"if$=\"
	.align 4, 0
	.int  1b
	.equ	ifsegmentsequalCFA, .
	");


asm	(" .int dupCFA ");
				/* ( len )				*/
asm	(" .int oneCFA ");

asm	(" .int plusCFA ");
			/* ( downcounter )			*/
asm	(" .int subCFA	");

 asm	(" .int p2dupCFA ");
			/* ( downcounter )   (P str str		*/

asm	("	.equ stringsequal_one, .	");


	asm	(" .int onbitsCFA ");
		/* ( dc -1		(P str str	*/
	asm	(" .int plusCFA ");
		/* ( dc-1		(P str str	*/
	asm	(" .int dupCFA ");
			/* ( dc-1 dc-1		(P str str	*/
	asm	(" .int noCFA ");
			/* ( dc-1		(P str str	*/
		asm	(" .int	 .  + 10  *4  ");

	asm	(" .int ifcontentsequalCFA ");
    /* ( dc-1 flag		(P str str	*/
	asm	(" .int subCFA	");

	 asm	(" .int p2plussCFA ");
		/* ( dc-1 flag		(P str' str'	*/
asm	(" .int yesCFA ");
				/* ( dc-1		(P str' str'	*/
asm	("	.int stringsequal_one	");


	asm	(" .int twopdropCFA ");
		/* ( dc-1		(P.)		*/
	asm	(" .int dropCFA ");
		/* ( )					*/
	asm	(" .int falseCFA ");
		/* ( false)				*/
	asm	(" .int ReturnCFA ");


		asm	(" .int twopdropCFA ");
	/* ( false		(P )		*/
		asm	(" .int dropCFA ");
	/* ( )					*/
		asm	(" .int trueCFA ");
	/* ( true)				*/
		asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============		*/

/* ifbyte$=	(P beg.	 beg. ||| --- ) ( len ||| --- flag )  255 max len */
  
asm	("
	1:
	.byte 8
	.byte 0
	.byte 0
	.byte 0
	.ascii \"ifbyte$=\"
	.align 4, 0
	.int  0b
	.equ	ifbytesegmentsequalCFA, .
	");


asm	(" .int dupCFA ");
				/* ( len )				*/
asm	(" .int sTOrCFA ");

asm	(" .int bytesCFA ");
		/* bytes context */
asm	(" .int oneCFA ");
				/* ( len 1 )				*/
asm	(" .int plusCFA ");
			/* ( downcounter )			*/
asm	(" .int subCFA	");

 asm	(" .int p2dupCFA ");
			/* ( downcounter )   (P str str		*/

asm	("	.equ bytestringsequal_one, .	");


	asm	(" .int onbitsCFA ");
		/* ( dc -1		(P str str	*/
	asm	(" .int plusCFA ");
		/* ( dc-1		(P str str	*/
	asm	(" .int dupCFA ");
			/* ( dc-1 dc-1		(P str str	*/
	asm	(" .int noCFA ");
			/* ( dc-1		(P str str	*/
		asm	(" .int	 .  + 11 *4  ");

	asm	(" .int ifcontentsequalCFA ");
    /* ( dc-1 flag		(P str str	*/
	asm	(" .int subCFA	");

	 asm	(" .int p2plussCFA ");
		/* ( dc-1 flag		(P str' str'	*/
asm	(" .int yesCFA ");
				/* ( dc-1		(P str' str'	*/
asm	("	.int bytestringsequal_one	");


asm	(" .int twopdropCFA ");
			/* ( dc-1		(P.)		*/
asm	(" .int rTOsCFA ");

asm	(" .int dropCFA ");

asm	(" .int falseCFA ");
			/* ( false)				*/
asm	(" .int ReturnCFA ");


		asm	(" .int twopdropCFA ");
	/* ( dc-1=0		(P )		*/
		asm	(" .int rTOsCFA ");

		asm	(" .int dropCFA ");

		asm	(" .int trueCFA ");
	/* ( true)				*/
		asm	(" .int ReturnCFA ");




 /* ttttttttttttttttttttttt =========== */

/* ifhex		( --- flag ) (P char ||| )   */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 1
	.ascii \"ifhex\"
	.align 4, 0
	.int  1b
	.equ	ifhexCFA, .
	");

asm	(" .int fetchCFA ");
		       /*   (...   char	    (P... char |||	  */
asm	(" .int maskbyteCFA ");
		       /*   (...  charbyte  (P... char |||
					CHARSIZE ASSUMPTION */
asm	(" .int subCFA	");

asm	(" .int hexdigitsCFA ");
			/*   (...  charbyte  (P... char ||| array  */
asm	(" .int pluspCFA ");
			/*   (...	     (P... char ||| index  */
asm	(" .int fetchCFA ");
			/*   (... flag	     (P... char ||| index  */
asm	(" .int pdropCFA ");

asm	(" .int maskbyteCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============		*/

/* include 	"incl. relative_pathname "	*/
  
asm	("
	1:
	.byte 7
	.byte 0
	.byte 0
	.byte 0
	.ascii \"include\"
	.align 4, 0
	.int  0b
	.equ	includeCFA, .
	");

asm	(" .int zeroCFA ");

asm	(" .int zeroCFA ");

asm	(" .int subCFA	");

  asm	(" .int openCFA ");

asm	(" .int subCFA	");

  asm	(" .int inputCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============		*/

/* input		( FD --- ) interpret input on FD til exhausted */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"input\"
	.align 4, 0
	.int  1b
	.equ	inputCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int sourceCFA ");

asm	(" .int storeCFA ");

asm	(" .int pdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int refillCFA ");
		/* (--was_some) @source !this !buffer !epa */ 
		asm	(" .int noCFA ");

		asm	(" .int	 .  + 9 *4  ");

asm	("	.equ while_input_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int tokenCFA ");
		/* (EOI)			*/
		asm	(" .int yesCFA ");
		/* ()				*/
		asm	(" .int	 .  + 5 *4  ");

	asm	(" .int subCFA	");

	  asm	(" .int interpretCFA ");
		/*   user context  ????????	*/
asm	(" .int ellCFA ");

asm	("	.int while_input_one	");

		asm	(" .int zeroCFA ");

		asm	(" .int subCFA	");

		  asm	(" .int sourceCFA ");

		asm	(" .int storeCFA ");

		asm	(" .int pdropCFA ");

		asm	(" .int subCFA	");

		  asm	(" .int refillCFA ");

		asm	(" .int dropCFA ");

		asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* interpret	interpret one token. Unknown stack effects. EVAL basically. */
  
asm	("
	1:
	.byte 9
	.byte 0
	.byte 0
	.byte 0
	.ascii \"interpret\"
	.align 4, 0
	.int  0b
	.equ	interpretCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int ifwordCFA ");
			/* (P HNC/NULL ) ( flag)		*/
	asm	(" .int noCFA ");
			/* (P HNC/NULL ) ()			*/
	asm	(" .int	 .  + 4 *4  ");

asm	(" .int subCFA	");
	
 asm	(" .int drowodCFA ");
			/* user context ??????????????????	*/
asm	(" .int ReturnCFA ");

	asm	(" .int subCFA	");

	 asm	(" .int ifabbreviationCFA ");
	/* (P HNC/NULL ) ( flag)	*/
		asm	(" .int noCFA ");
		/* (P HNC/NULL ) ()			*/
		asm	(" .int	 .  + 4 *4  ");

	asm	(" .int subCFA	");
	
	 asm	(" .int drowodCFA ");
		/* user context ??????????????????	*/
	asm	(" .int ReturnCFA ");

		asm	(" .int subCFA	");
 
		 asm	(" .int ifhexnumberCFA ");

			asm	(" .int noCFA ");

			asm	(" .int	 .  + 4 *4  ");

		asm	(" .int subCFA	");

		  asm	(" .int rebmunxehCFA ");

		asm	(" .int ReturnCFA ");

asm	(" .int subCFA	");
 asm	(" .int crCFA ");

asm	(" .int subCFA	");

  asm	(" .int tbCFA ");

asm	(" .int subCFA	");
 
  asm	(" .int toklenCFA ");

asm	(" .int fetchCFA ");

asm	(" .int pdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int printstringCFA ");

asm	(" .int subCFA	");

  asm	(" .int beepCFA ");

asm	(" .int pdropCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============		*/

/* initresolve	init branch resolver stacks */
  
asm	("
	0:
	.byte 11
	.byte 0
	.byte 0
	.byte 1
	.ascii \"initresolve\"
	.align 4, 0
	.int  1b
	.equ	initresolveCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int branchstackCFA ");
		/* (P &bs)		*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
			/* (P &bs &bsi)		*/
asm	(" .int pstoreCFA ");
			/* (P &bs &bsi)		*/
asm	(" .int twopdropCFA ");

asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
		/* (P &bs)		*/
asm	(" .int subCFA	");

 asm	(" .int tsiCFA ");
			/* (P &bs &bsi)		*/
asm	(" .int pstoreCFA ");
			/* (P &bs &bsi)		*/
asm	(" .int twopdropCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* link		!dp   !latest	  link a word to the dictionary */
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"link\"
	.align 4, 0
	.int  0b
	.equ	linkCFA, .
	");


asm	(" .int subCFA	");

 asm	(" .int latestCFA ");
		/*	(P  &latest)			*/
asm	(" .int pfetchCFA ");
		/*	(P  latest)			*/
asm	(" .int subCFA	");

 asm	(" .int appendpCFA ");
		/*	(P  latest)   !dp !HERE		*/
asm	(" .int pdropCFA ");
		/*	(P)				*/
asm	(" .int subCFA	");

 asm	(" .int pvarCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int subCFA	");

 asm	(" .int latestCFA ");

asm	(" .int pstoreCFA ");

asm	(" .int twopdropCFA ");

asm	(" .int ReturnCFA ");



  /* tttttttttttttttttttttt===========				*/

/* -		( a b ---  a-b )  minus		*/
  
asm	("
	0:
	.byte 1
	.byte 0
	.byte 0
	.byte 0
	.ascii \"-\"
	.align 4, 0
	.int  1b
	.equ	minusCFA, .
	");

asm	(" .int negateCFA ");

asm	(" .int plusCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* name		"name newname "	   part of :			*/
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"name\"
	.align 4, 0
	.int  0b
	.equ	headerCFA, .
	");

			/* ENDIANISM THANG */
asm	(" .int sTOrCFA ");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int subCFA	");

 asm	(" .int pvarCFA ");

asm	(" .int pstoreCFA ");
		/* link uses the value in pvar. Fugly.		*/
asm	(" .int twopdropCFA ");

asm	(" .int subCFA	");

 asm	(" .int tokenCFA ");
		/* (flag)			*/
asm	(" .int dropCFA ");
		/* ()				*/
asm	(" .int bytes_bdCFA ");

asm	(" .int subCFA	");

 asm	(" .int toklenCFA ");
		/*		(P  &toklen)			*/
asm	(" .int fetchCFA ");
		/*		(P  &toklen)  ( toklen )	*/
asm	(" .int subCFA	");

 asm	(" .int appendCFA ");
		/*		(P  &toklen)			*/
asm	(" .int pdropCFA ");
		/*		(P )				*/
asm	(" .int zeroCFA ");
		/* (0)						*/
asm	(" .int subCFA	");

 asm	(" .int appendCFA ");
		/* ()						*/
asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 0	");
		/* (0)		non-atomic			*/
asm	(" .int subCFA	");

asm	(" .int appendCFA ");
		/* ()						*/
asm	(" .int zeroCFA ");
		/* (0)						*/
asm	(" .int subCFA	");

 asm	(" .int appendCFA ");
		/* ()						*/
asm	(" .int rTOsCFA ");

asm	(" .int subCFA	");

 asm	(" .int tbCFA ");
			/* ()		(P tb )		    */
asm	(" .int subCFA	");

 asm	(" .int toklenCFA ");
		/* ()		(P tb	&toklen)      */
asm	(" .int fetchCFA ");
		/* (len)	(P itb	&toklen)      */
asm	(" .int pdropCFA ");

asm	(" .int dupCFA ");
			/* (len len)	(P itoken)		*/
asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (len len)	(P itoken &dp)		*/
asm	(" .int pfetchCFA ");
		/* (len len)	(P itoken dp)		*/
asm	(" .int moveCFA ");
		/* (len)	(P itoken dp)		*/
asm	(" .int pluspCFA ");
		/* ()		(P itoken dp')		*/
asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* ()		(P itoken dp' &dp)	*/
asm	(" .int pstoreCFA ");
		/* ()		(P itoken dp' &dp)	*/
asm	(" .int pdropCFA ");

asm	(" .int alignCFA ");
		/* ()		(P itoken dp_al)	*/
asm	(" .int subCFA	");

 asm	(" .int dpCFA ");

asm	(" .int pstoreCFA ");

asm	(" .int twopdropCFA ");

asm	(" .int twopdropCFA ");
			SUSPECT
   asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* open		( flags perms --- FD|err )  "open  filename " @!tb	*/
  
asm	("
	0:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"open\"
	.align 4, 0
	.int  1b
	.equ	openCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int TO_CCFA ");
		/* (P tb )	(flags perms)		*/
asm	(" .int sys_openCFA ");
		/* (P tb )	(FD)			*/
asm	(" .int pdropCFA ");
		/* (P)		(FD)			*/
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============				*/

/* postpone 	"postpone bla " CFA of bla will be appended by child 
		SUSPECT
		DOESN'T WORK FROM AN INCLUDE FILE	*/
  
asm	("
	1:
	.byte 8
	.byte 1
	.byte 0
	.byte 0
	.ascii \"postpone\"
	.align 4, 0
	.int  0b
	.equ	postponeCFA, .
	");

asm	(" .int subCFA	");
 
  asm	(" .int tickCFA ");
		/* (P HNCbla|null )	(flag)	*/
asm	(" .int dropCFA ");
		/* assume word			*/
asm	(" .int TOcodeCFA ");
		/* (P CFAbla )			*/
asm	(" .int pliteralCFA ");
		
asm	(" .int pliteralCFA ");
		/* (P CFAbla CFAplit ) 		*/		
asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");
		/* (P CFAbla)			*/
asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");
		/* (P)				*/
asm	(" .int pliteralCFA ");

asm	(" .int subCFA	");

asm	(" .int subCFA	");
			/* (P CFAsub)			*/
  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");
		/* (P)				*/
asm	(" .int pliteralCFA ");

asm	(" .int appendpCFA ");
		/* (P CFA&p )			*/
asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");
		
asm	(" .int pdropCFA ");
		/* (P)				*/
asm	(" .int pliteralCFA ");
		
asm	(" .int pdropCFA ");
		/* (P CFApdrop )		*/
asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");
		
asm	(" .int pdropCFA ");
		/* (P)				*/
asm	(" .int trueCFA ");
		/* ( true ) 			*/
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============				*/

/* p2dup		(P a b ||| --- a b ) dup 2 pointers in-phase	*/
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"p2dup\"
	.align 4, 0
	.int  1b
	.equ	p2dupCFA, .
	");

asm	(" .int poverCFA ");

asm	(" .int poverCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* p2+s		(P  ptr	 ptr  --- ptr+s	 ptr+s )		*/
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"p2+s\"
	.align 4, 0
	.int  0b
	.equ	p2plussCFA, .
	");

asm	(" .int pplussCFA ");

asm	(" .int pdropCFA ");

asm	(" .int pplussCFA ");

asm	(" .int pupCFA ");

asm	(" .int ReturnCFA ");


/*		see x for Forth "." until H3sm gets BASE			*/


  /* ttttttttttttttttttttttt============			*/
/* .p		print pointer  non-destructively		*/
  
asm	("
	0:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \".p\"
	.align 4, 0
	.int  1b
	.equ	printpCFA, .
	");

asm	(" .int pdupCFA ");

asm	(" .int pTOCFA ");
		/*  ( --- p ) 4 bytes moved regardless		*/
asm	(" .int printhexCFA ");
	/*  ( --- )					*/
asm	(" .int ReturnCFA ");



   /* ttttttttttttttttttttttt============			*/
/* .r		print TOR nondestructively			*/
  
asm	("
	1:
	.byte 2
	.byte 0
	.byte 0
	.byte 0
	.ascii \".r\"
	.align 4, 0
	.int  0b
	.equ	printrCFA, .
	");

asm	(" .int rpcopyCFA ");

asm	(" .int printhexCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/
/* .string	(P begin  ||| ) ( length --- ) print a string	*/
  
asm	("
	0:
	.byte 7
	.byte 0
	.byte 0
	.byte 0
	.ascii \".string\"
	.align 4, 0
	.int  1b
	.equ	printstringCFA, .
	");

asm	(" .int pdupCFA ");

asm	(" .int TOrCFA ");

asm	("	.equ prints_one, .	");

	asm	(" .int fetchCFA ");

	asm	(" .int emitCFA ");

	asm	(" .int minusaskrCFA ");

	asm	(" .int pplusbCFA ");

	asm	(" .int yesCFA ");

asm	("	.int prints_one	");

asm	(" .int pdropCFA ");

asm	(" .int rdropCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* .chars		(P begin  ||| ) ( length --- ) prt 126 > chars > 32*/
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \".chars\"
	.align 4, 0
	.int  0b
	.equ	printcharsCFA, .
	");

asm	(" .int TOrCFA ");

asm	(" .int pdupCFA ");
			/* (P str) (R len )		*/
asm	("	.equ printchars_one, .	");

	asm	(" .int fetchCFA ");
		/* ( c )			*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 32	");
		/* ( c blank)			*/
	asm	(" .int maxCFA ");
			/* ( c|blank)			*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 126	");
		/* ( c|blank  ~)		*/
	asm	(" .int minCFA ");
			/* ( c|blank|~)			*/
	asm	(" .int emitCFA ");
		/* ()				*/
	asm	(" .int minusaskrCFA ");
		/* ( flag )		(R n-1) */
	asm	(" .int pplusbCFA ");
		/*  ( flag ) (P str')	(R n-1) */
	asm	(" .int yesCFA ");
			/*  () (P str')	  (R n-1)	*/
asm	("	.int printchars_one	");


asm	(" .int rdropCFA ");
			/*  () (P str')			*/
asm	(" .int pdropCFA ");

asm	(" .int ReturnCFA ");



 /* |||||||||||||||||||||||||||||||||============		*/

/* .name		(P   HNC ||| )	print a word's H3m name */
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \".name\"
	.align 4, 0
	.int  1b
	.equ	printnameCFA, .
	");

asm	(" .int pdupCFA ");

asm	(" .int fetchCFA ");

asm	(" .int maskbyteCFA ");

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 4	");

asm	(" .int pluspCFA ");

asm	(" .int onbitsCFA ");

asm	(" .int plusCFA ");

asm	(" .int subCFA	");

  asm	(" .int printstringCFA ");

asm	(" .int pdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int blankCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/
/* resolve_one	resolve one branch in thread being threaded */
  
asm	("
	1:
	.byte 11
	.byte 0
	.byte 0
	.byte 1
	.ascii \"resolve_one\"
	.align 4, 0
	.int  0b
	.equ	resolve_oneCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int bsiCFA ");
	/* (P  &bsi)				*/
asm	(" .int pfetchCFA ");
		/* (P  bsi )				*/
asm	(" .int pminuscCFA ");
		/* (P  bsi-)				*/
asm	(" .int pdupCFA ");
		/* (P  bsi-  bsi-)			*/
asm	(" .int pfetchCFA ");
		/* (P  bsi-  targ.stck.addy)	pop	*/
asm	(" .int pfetchCFA ");
		/* (P  bsi-  targ.addy)			*/
asm	(" .int pswapCFA ");
		/* (P  targ.addy bsi-)			*/
asm	(" .int pminuscCFA ");
		/* (P  targ.addy bsi--)			*/
asm	(" .int pdupCFA ");
		/* (P  targ.addy bsi-- bsi--)		*/
asm	(" .int pfetchCFA ");
		/* (P  targ.addy bsi-- poke.loc )	*/
asm	(" .int pswapCFA ");
		/* (P targ.addy poke.loc bsi--)		*/
asm	(" .int pTOrCFA ");
		/* (P targ.addy poke.loc bsi--) (R bsi--)	*/
asm	(" .int pstoreCFA ");
		/*  (P targ.addy poke.loc) (R bsi--) RESOLVED!*/
asm	(" .int twopdropCFA ");
		/* (P ) (R bsi--)			*/
asm	(" .int rTOpCFA ");
		/* (P bsi--)				*/
asm	(" .int subCFA	");

  asm	(" .int bsiCFA ");
	/* (P bsi-- &bsi)			*/
asm	(" .int pstoreCFA ");

asm	(" .int twopdropCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* refill		( --- got ) @source !this !epa !buffer   buffered more input*/

  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"refill\"
	.align 4, 0
	.int  1b
	.equ	refillCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int sourceCFA ");
		/* (p &source)  	 			*/
asm	(" .int fetchCFA ");
		/* (P)			(fd)			*/
asm	(" .int pdropCFA ");

asm	(" .int subCFA	");
	
  asm	(" .int bufferCFA ");
		/* (P buffer)		(fd)			*/
asm	(" .int readCFA ");
		/* (P buffer)		( got)			*/ 
asm	(" .int dupCFA ");
			/* (P buffer)		( got got)		*/
asm	(" .int subCFA	");
	
  asm	(" .int thispointerCFA ");
	/* (P buffer tp)	( got got)		*/	
asm	(" .int pstoreCFA ");
		/* (P buffer->tp)	( got got )		*/
asm	(" .int pdropCFA ");
		/* (P buffer)		( got got )		*/
asm	(" .int pluspCFA ");
		/* (P buffer+)		( got )			*/
asm	(" .int subCFA	");

  asm	(" .int epaCFA ");
		/* (P buffer+ &epa)	( got )			*/
asm	(" .int pstoreCFA ");
		/* (P buffer+->&epa)	( got )			*/
asm	(" .int twopdropCFA ");
		/* (P)			( got )			*/
asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* resolve	(--- true) @bs @ts, thread all pending branches THREADer
			resolve uses bs decr.-pop style		*/
  
asm	("
	1:
	.byte 7
	.byte 1
	.byte 0
	.byte 0
	.ascii \"resolve\"
	.align 4, 0
	.int  0b
	.equ	resolveCFA, .
	");


asm	("	.equ resolve_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int branchstackCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int bsiCFA ");
		/* (P  &bs &bsi)		*/
	asm	(" .int pfetchCFA ");
		/* (P  &bs bsi )		*/
	asm	(" .int gapCFA ");
			/* (P  &bs bsi )  (gap)		*/
	asm	(" .int twopdropCFA ");
		/* (gap)			*/
	asm	(" .int noCFA ");
			/* a go# is pending		*/
		asm	(" .int	 .  + 5 *4  ");
	/* no branches pending		*/
	asm	(" .int subCFA	");
			/* resolve a pending branch	*/
	  asm	(" .int resolve_oneCFA ");
	/* ()				*/
asm	(" .int ellCFA ");

asm	("	.int resolve_one	");

		asm	(" .int ReturnCFA ");



  /* tttttttttttttttttttttttttttttttttttttttt============ */

/* rebmunxeh	(  --- #_in_user_data_stack_context ) The Dreaded Rebmunxeh. */
  
asm	("
	0:
	.byte 9
	.byte 0
	.byte 0
	.byte 1
	.ascii \"rebmunxeh\"
	.align 4, 0
	.int  1b
	.equ	rebmunxehCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int USizeCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int bumpCFA ");

asm	(" .int pTOsCFA ");

asm	(" .int dropCFA ");

asm	(" .int subCFA	");

  asm	(" .int toklenCFA ");

asm	(" .int fetchCFA ");
			/*   (len)				*/
asm	(" .int pdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int tbCFA ");
			/*	(len)		(P tb)		*/	
asm	(" .int zeroCFA ");
			/*   ( len  accum0	(P  bt		*/
asm	("	.equ rebmunx_one, .	");

	asm	(" .int overCFA ");
		/*   ( len accum len	(P char		*/
	asm	(" .int noCFA ");
			/*   ( len accum	(P char		*/
		asm	(" .int	 .  + 15 *4  ");

	asm	(" .int literalCFA ");
		/*   ( len  accum  4	(P char		*/
	asm	(" .int 4	");
			/*   ( len  accum<	(P char		*/
	asm	(" .int 4	");

	asm	(" .int upshiftCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int asciiTOdigitCFA ");
	/*  ( len  accum< dig	(P char		*/
	asm	(" .int ORCFA ");
			/*  ( len  accum<+	(P char		*/
	asm	(" .int swapCFA ");
		/*  ( accum  cnt	(P char		*/
	asm	(" .int onbitsCFA ");
		/*  ( accum  cnt   -1	(P char		*/
	asm	(" .int plusCFA ");
		/*  ( accum  cntdn			*/
	asm	(" .int swapCFA ");
		/*  ( cntdn  accum			*/
	asm	(" .int pplusbCFA ");
		/*  ( cntdn  accum	(P char'	*/
asm	(" .int ellCFA ");

  asm	("	.int rebmunx_one	");

				/*  ( len accum (P char			*/
		asm	(" .int pdropCFA ");
	/*  ( len accum				*/
		asm	(" .int swapCFA ");
	/*  ( accu  len				*/
		asm	(" .int dropCFA ");
	/*  ( accu ) leave this in user context */

		asm	(" .int cells_bdCFA ");

		asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============			*/

/* ;		( --- stop_THREADer) set a target, resolve, & a return. */
  
asm	("
	1:
	.byte 1
	.byte 1
	.byte 0
	.byte 0
	.ascii \";\"
	.align 4, 0
	.int  0b
	.equ	semicolonCFA, .
	");

				/* (O) for to# = ;		*/
asm	(" .int subCFA	");

 asm	(" .int targetCFA ");
		/* (P &dp)				*/
asm	(" .int dropCFA ");
		/* (P dp)				*/
asm	(" .int subCFA	");
			/* resolve branches			*/
  asm	(" .int resolveCFA ");

asm	(" .int pliteralCFA ");
	/* THREAD  a return				*/
  asm	(" .int ReturnCFA ");
		/* (P )		( ReturnCFA )		*/
asm	(" .int subCFA	");

 asm	(" .int appendpCFA ");
		 /* (P )				*/
asm	(" .int pdropCFA ");
		SUSPECT
asm	(" .int falseCFA ");
		/* tell : to stop threading		*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* spaces		( N --- )  emit N spaces	*/
  
asm	("
	0:
	.byte 6
	.byte 0
	.byte 0
	.byte 0
	.ascii \"spaces\"
	.align 4, 0
	.int  1b
	.equ	spacesCFA, .
	");

asm	(" .int TOrCFA ");

asm	("	.equ spaces_one, .	");

      asm	(" .int subCFA	");

	asm	(" .int blankCFA ");

	asm	(" .int minusaskrCFA ");

	asm	(" .int yesCFA ");

	asm	("	.int spaces_one	");

asm	(" .int rdropCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt ============	*/

/* step		( --- EOI) true if End Of Input on this		*/
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"step\"
	.align 4, 0
	.int  0b
	.equ	stepCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int thispointerCFA ");

asm	(" .int pdupCFA ");
		/* increment *thispointer	*/
asm	(" .int pfetchCFA ");

asm	(" .int pplusbCFA ");

asm	(" .int pswapCFA ");

asm	(" .int pstoreCFA ");

asm	(" .int twopdropCFA ");

asm	(" .int subCFA	");

  asm	(" .int ifpendingCFA ");
		/* is there more buffer?  */
	asm	(" .int yesCFA ");

	asm	(" .int	 .  + 6 *4  ");

asm	(" .int subCFA	");

  asm	(" .int refillCFA ");
		/* refill checks for more input */
asm	(" .int flagCFA ");

asm	(" .int aintCFA ");

asm	(" .int ReturnCFA ");

	asm	(" .int falseCFA ");

	asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============			 */

/* snug		( --- EOI) eat whitespace. snug+finish=token. Buffered. */
  
asm	("
	0:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"snug\"
	.align 4, 0
	.int  1b
	.equ	snugCFA, .
	");


asm	("	.equ while_white_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int thispointerCFA ");
		/*  (P str )			*/
	asm	(" .int pfetchCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int ifcharCFA ");
			/*  ( --- ischar) does a pdrop	*/
	asm	(" .int noCFA ");

	asm	(" .int	 .  + 3 *4  ");

		asm	(" .int falseCFA ");

		    asm	(" .int ReturnCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int stepCFA ");
			/* ( --- EOI )			*/
	asm	(" .int noCFA ");

		asm	(" .int	 .  + 3 *4  ");

	asm	(" .int trueCFA ");

	    asm	(" .int ReturnCFA ");

asm	(" .int ellCFA ");

   asm	("	.int while_white_one	");



  /* tttttttttttttttttttttttttttttttttttttttttttttttttttt============ */

/* this		( --- current_input_buffer_byte ) interpreter part. */
  
asm	("
	1:
	.byte 4
	.byte 0
	.byte 0
	.byte 0
	.ascii \"this\"
	.align 4, 0
	.int  0b
	.equ	thisCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int thispointerCFA ");

asm	(" .int pfetchCFA ");

asm	(" .int fetchCFA ");

asm	(" .int pdropCFA ");

asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttttttttttttttttttttttt============ */

/* token		( --- End_Of_Input ) ?!tb !this		*/
  
asm	("
	0:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"token\"
	.align 4, 0
	.int  1b
	.equ	tokenCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int snugCFA ");
		/* (EOI)		*/
	asm	(" .int yesCFA ");
		/* ()			*/
	asm	(" .int	 .  + 6 *4  ");

asm	(" .int subCFA	");

  asm	(" .int finishCFA ");
		/* (EOI)		*/
asm	(" .int dropCFA ");
		/* ()			*/
asm	(" .int falseCFA ");

asm	(" .int ReturnCFA ");

	asm	(" .int trueCFA ");

	asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* to:		( --- keep_THREADer) set branch to resolve to start of word
				stack and make dict. room for br. */
  
asm	("
	1:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"to:\"
	.align 4, 0
	.int  0b
	.equ	tocolonCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (P &dp)					*/
asm	(" .int pfetchCFA ");
		/* (P dp)					*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
	/* (P dp  &bsi)					*/
asm	(" .int pfetchCFA ");
		/* (P dp   bsi)					*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi)		!branchstackcell	*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi')				*/
asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
	/* (P dp   bsi' &ts0)	target spec 0		*/
asm	(" .int pswapCFA ");
		/* (P dp   &ts0 bsi')				*/
asm	(" .int pstoreCFA ");
		/* (P dp   &ts0 bsi')	!branchstackcell	*/
asm	(" .int pswapCFA ");
		/* (P dp   bsi' &ts0)				*/
asm	(" .int pdropCFA ");
		/* (P dp   bsi')				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi'')				*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
   /* (P dp   bsi'' &bsi)				*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi'' &bsi)	 double-incr index	*/
asm	(" .int twopdropCFA ");
		/* (P dp )					*/
asm	(" .int ppluscCFA ");
		/* (P dp')					*/
asm	(" .int subCFA	");

  asm	(" .int dpCFA ");
		/* (P dp' &dp)					*/
asm	(" .int pstoreCFA ");
		/* (P dp' &dp)		dp+c			*/
asm	(" .int twopdropCFA ");
		/* (P )						*/
asm	(" .int trueCFA ");
		/* (true)	say "keep going"		*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* to1		( --- keep_THREADing) set branch to first (O) to resolve
				stack and make dict. room for br. */
  
asm	("
	0:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"to1\"
	.align 4, 0
	.int  1b
	.equ	to1CFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (P &dp)					*/
asm	(" .int pfetchCFA ");
		/* (P dp)					*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
	/* (P dp  &bsi)					*/
asm	(" .int pfetchCFA ");
		/* (P dp   bsi)					*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi)		!branchstackcell	*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi')				*/
asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
	/* (P dp   bsi' &ts0)				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi' &ts1)	target spec 1		*/
asm	(" .int pswapCFA ");
		/* (P dp   &ts1 bsi')				*/
asm	(" .int pstoreCFA ");
		/* (P dp   &ts1 bsi')	!branchstackcell	*/
asm	(" .int pswapCFA ");
		/* (P dp   bsi' &ts1)				*/
asm	(" .int pdropCFA ");
		/* (P dp   bsi')				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi'')				*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
   /* (P dp   bsi'' &bsi)				*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi'' &bsi)	 double-incr index	*/
asm	(" .int twopdropCFA ");
		/* (P dp )					*/
asm	(" .int ppluscCFA ");
		/* (P dp')					*/
asm	(" .int subCFA	");

  asm	(" .int dpCFA ");
		/* (P dp' &dp)					*/
asm	(" .int pstoreCFA ");
		/* (P dp' &dp)		dp+c			*/
asm	(" .int twopdropCFA ");
		/* (P )						*/
asm	(" .int trueCFA ");
		/* (true)	keep threading			*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* to2		( --- keep_THReding) set branch to second (O) to resolve
				stack and make dict. room for br. */
  
asm	("
	1:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"to2\"
	.align 4, 0
	.int  0b
	.equ	to2CFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (P &dp)					*/
asm	(" .int pfetchCFA ");
		/* (P dp)					*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
	/* (P dp  &bsi)					*/
asm	(" .int pfetchCFA ");
		/* (P dp   bsi)					*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi)		!branchstackcell	*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi')				*/
asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
	/* (P dp   bsi' &ts0)				*/
asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");
		/* (P dp   bsi' &ts2)	target spec 2		*/
asm	(" .int pswapCFA ");
		/* (P dp   &ts2 bsi')				*/
asm	(" .int pstoreCFA ");
		/* (P dp   &ts2 bsi')	!branchstackcell	*/
asm	(" .int pswapCFA ");
		/* (P dp   bsi' &ts2)				*/
asm	(" .int pdropCFA ");
		/* (P dp   bsi')				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi'')				*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
   /* (P dp   bsi'' &bsi)				*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi'' &bsi)	 double-incr index	*/
asm	(" .int twopdropCFA ");
		/* (P dp )					*/
asm	(" .int ppluscCFA ");
		/* (P dp')					*/
asm	(" .int subCFA	");

  asm	(" .int dpCFA ");
		/* (P dp' &dp)					*/
asm	(" .int pstoreCFA ");
		/* (P dp' &dp)		dp+c			*/
asm	(" .int twopdropCFA ");
		/* (P )						*/
asm	(" .int trueCFA ");
		/* (true)	keep threading			*/
asm	(" .int ReturnCFA ");



/* to3		( --- keep_THREDING) as above, reserve a cell
				stack and make dict. room for br. */
  
asm	("
	0:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"to3\"
	.align 4, 0
	.int  1b
	.equ	to3CFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (P &dp)					*/
asm	(" .int pfetchCFA ");
		/* (P dp)					*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
	/* (P dp  &bsi)					*/
asm	(" .int pfetchCFA ");
		/* (P dp   bsi)					*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi)		!branchstackcell	*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi')				*/
asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
	/* (P dp   bsi' &ts0)				*/
asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");
		/* (P dp   bsi' &ts3)	target spec 3		*/
asm	(" .int pswapCFA ");
		/* (P dp   &ts0 bsi')				*/
asm	(" .int pstoreCFA ");
		/* (P dp   &ts0 bsi')	!branchstackcell	*/
asm	(" .int pswapCFA ");
		/* (P dp   bsi' &ts0)				*/
asm	(" .int pdropCFA ");
		/* (P dp   bsi')				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi'')				*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
   /* (P dp   bsi'' &bsi)				*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi'' &bsi)	 double-incr index	*/
asm	(" .int twopdropCFA ");
		/* (P dp )					*/
asm	(" .int ppluscCFA ");
		/* (P dp')					*/
asm	(" .int subCFA	");

  asm	(" .int dpCFA ");
		/* (P dp' &dp)					*/
asm	(" .int pstoreCFA ");
		/* (P dp' &dp)		dp+c			*/
asm	(" .int twopdropCFA ");
		/* (P )						*/
asm	(" .int trueCFA ");
		/* (true)	keep threading			*/
asm	(" .int ReturnCFA ");



/* to4		( --- keep_THREDING) as above. Max (O) + 1 is at ;
				stack and make dict. room for br. */
  
asm	("
	1:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"to4\"
	.align 4, 0
	.int  0b
	.equ	to4CFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
			/* (P &dp)					*/
asm	(" .int pfetchCFA ");
		/* (P dp)					*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
		/* (P dp  &bsi)					*/
asm	(" .int pfetchCFA ");
		/* (P dp   bsi)					*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi)		!branchstackcell	*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi')				*/
asm	(" .int subCFA	");

 asm	(" .int targetstackCFA ");
	/* (P dp   bsi' &ts0)				*/
asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");

asm	(" .int ppluscCFA ");
		/* (P dp   bsi' &ts0)	target spec 4		*/
asm	(" .int pswapCFA ");
		/* (P dp   &ts0 bsi')				*/
asm	(" .int pstoreCFA ");
		/* (P dp   &ts0 bsi')	!branchstackcell	*/
asm	(" .int pswapCFA ");
		/* (P dp   bsi' &ts0)				*/
asm	(" .int pdropCFA ");
		/* (P dp   bsi')				*/
asm	(" .int ppluscCFA ");
		/* (P dp   bsi'')				*/
asm	(" .int subCFA	");

 asm	(" .int bsiCFA ");
		/* (P dp   bsi'' &bsi)				*/
asm	(" .int pstoreCFA ");
		/* (P dp   bsi'' &bsi)	 double-incr index	*/
asm	(" .int twopdropCFA ");
		/* (P dp )					*/
asm	(" .int ppluscCFA ");
		/* (P dp')					*/
asm	(" .int subCFA	");

  asm	(" .int dpCFA ");
		/* (P dp' &dp)					*/
asm	(" .int pstoreCFA ");
		/* (P dp' &dp)		dp+c			*/
asm	(" .int twopdropCFA ");
		/* (P )						*/
asm	(" .int trueCFA ");
		/* (true)	keep threading			*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* tab		emit a tab, ASCII 9		*/
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 0
	.ascii \"tab\"
	.align 4, 0
	.int  1b
	.equ	tabCFA, .
	");

asm	(" .int literalCFA ");

 asm	(" .int 4	");

 asm	(" .int 9	");

asm	(" .int emitCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* (O)		target ( --- keep_THREDING) set branch target 1 2 3 or 4*/
  
asm	("
	1:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"(O)\"
	.align 4, 0
	.int  0b
	.equ	targetCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int dpCFA ");
				/* (P &dp)				*/
asm	(" .int pfetchCFA ");
			/* (P dp)				*/
asm	(" .int subCFA	");

 asm	(" .int tsiCFA ");
			/* (P dp &tsi)				*/
asm	(" .int pfetchCFA ");
			/* (P dp tsi)				*/
asm	(" .int pstoreCFA ");
			/* (P dp tsi)	    target stashed	*/
asm	(" .int ppluscCFA ");
			/* (P dp tsi')				*/
asm	(" .int subCFA	");

 asm	(" .int tsiCFA ");
			/* (P dp tsi' &tsi)			*/
asm	(" .int pstoreCFA ");
			/* (P dp tsi' &tsi)  ts index incr'd	*/
asm	(" .int pdropCFA ");
			/* (P dp tsi')				*/
asm	(" .int twopdropCFA ");

asm	(" .int trueCFA ");
			/* (true)   keep threading		*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* threadtoken	( --- keep_going? )  called after "token"	*/
  
asm	("
	0:
	.byte 11
	.byte 0
	.byte 0
	.byte 1
	.ascii \"threadtoken\"
	.align 4, 0
	.int  1b
	.equ	threadtokenCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int ifwordCFA ");
	       /* (P ---  &HNC|null) ( --- flag )      */
asm	(" .int noCFA ");
			/* (P ---  &HNC|null) ()		*/
 asm	(" .int	 .  + 10 *4  ");

asm	(" .int subCFA	");

 asm	(" .int ifthreaderCFA ");
		/* (P ---  &HNC) ( --- flag )	       */
	asm	(" .int yesCFA ");
		/* (P ---  &HNC) ()		       */
	asm	(" .int	 .  + 63 *4  ");

asm	(" .int subCFA	");

 asm	(" .int appendwordCFA ");
		/* (P ---  &HNC) ()			*/
asm	(" .int trueCFA ");
		/* (P ---  &HNC) ( --- true)		*/
asm	(" .int pdropCFA ");
		/* (P) ( --- true)			*/
asm	(" .int ReturnCFA ");
		/* (P) ( --- true)			*/

asm	(" .int subCFA	");
			/* (P;NULL)				*/
 asm	(" .int ifabbreviationCFA ");
	/*				       */
	asm	(" .int noCFA ");
		/*  is it an abbrev. word?		*/
	asm	(" .int	 .  + 10 *4  ");
	/* no. not a word at all.	       */
asm	(" .int subCFA	");
			/* yes, so it's a word.			*/
 asm	(" .int ifthreaderCFA ");

	asm	(" .int yesCFA ");
		/* threader (immediate) word?		*/
	asm	(" .int	 .  + 50 *4  ");
	/* yes. go do it now.		       */
asm	(" .int subCFA	");
			/* no. thread it.			*/
 asm	(" .int appendwordCFA ");

asm	(" .int pdropCFA ");

asm	(" .int trueCFA ");

asm	(" .int ReturnCFA ");


asm	(" .int subCFA	");

 asm	(" .int ifhexnumberCFA ");

	asm	(" .int noCFA ");
 
	asm	(" .int	 .  + 12 *4  ");
	   
asm	(" .int subCFA	");

 asm	(" .int hexnumberCFA ");

asm	(" .int subCFA	");

 asm	(" .int appendlitCFA ");

asm	(" .int trueCFA ");

asm	(" .int ReturnCFA ");


asm	(" .int subCFA	");
			/* stub			*/
 asm	(" .int iffileCFA ");

	asm	(" .int noCFA ");

	asm	(" .int	 .  + 5 *4  ");

asm	(" .int subCFA	");

 asm	(" .int H3mCFA ");

asm	(" .int trueCFA ");

asm	(" .int ReturnCFA ");
	
	asm	(" .int subCFA	");

	  asm	(" .int beepCFA ");

	asm	(" .int subCFA	");

	 asm	(" .int toklenCFA ");

	asm	(" .int fetchCFA ");

	asm	(" .int pdropCFA ");

asm	(" .int subCFA	");
 
 asm	(" .int tbCFA ");
 

asm	(" .int subCFA	");
 asm	(" .int crCFA ");
 asm	(" .int subCFA	");
 asm	(" .int crCFA ");
 asm	(" .int subCFA	");
 asm	(" .int tabCFA ");
 asm	(" .int subCFA	");
 asm	(" .int tabCFA ");


asm	(" .int subCFA	");

asm	(" .int printstringCFA ");


asm	(" .int subCFA	");
 asm	(" .int crCFA ");
 asm	(" .int subCFA	");
 asm	(" .int crCFA ");
 asm	(" .int subCFA	");
 asm	(" .int crCFA ");


asm	(" .int pdropCFA ");

asm	(" .int ReturnCFA ");

			/* token is threader word */
asm	(" .int doHNCCFA ");

asm	(" .int ReturnCFA ");



  /* ttttttttttttttttttttttt============ */

/* threadloop	called by :. Build the body of a thread. Stops after ;.	*/
  
asm	("
	1:
	.byte 10
	.byte 0
	.byte 0
	.byte 1
	.ascii \"threadloop\"
	.align 4, 0
	.int  0b
	.equ	threadloopCFA, .
	");


asm	("	.equ threadtokens_one, .	");

	asm	(" .int subCFA	");

	 asm	(" .int tokenCFA ");

		asm	(" .int yesCFA ");

		asm	(" .int	 .  + 7 *4  ");

	asm	(" .int subCFA	");

	 asm	(" .int threadtokenCFA ");

		asm	(" .int noCFA ");

		asm	(" .int	 .  + 3 *4  ");
	/* if keep-going is false	*/
asm	(" .int ellCFA ");
				/* keep-going was true		*/
asm	("	.int threadtokens_one	");

		asm	(" .int ReturnCFA ");



  /* TTTTTTTTTTTTTTTTTTTTTTTT============ */

/* '		(P --- HNC|null)  (--- flag) "' wordname "  Forth tick   */
  
asm	("
	0:
	.byte 1
	.byte 0
	.byte 0
	.byte 0
	.ascii \"'\"
	.align 4, 0
	.int  1b
	.equ	tickCFA, .
	");

asm	(" .int subCFA	");

 asm	(" .int tokenCFA ");

		asm	(" .int yesCFA ");

		asm	(" .int	 .  + 13 *4  ");

asm	(" .int subCFA	");

 asm	(" .int ifwordCFA ");

	asm	(" .int noCFA ");

	asm	(" .int	 .  + 3 *4  ");

asm	(" .int trueCFA ");

asm	(" .int ReturnCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int ifabbreviationCFA ");

		asm	(" .int noCFA ");

		asm	(" .int	 .  + 3 *4  ");

	asm	(" .int trueCFA ");

	asm	(" .int ReturnCFA ");

		asm	(" .int falseCFA ");

		asm	(" .int ReturnCFA ");



  /* tttttttttttttttttttttttt============ */

/* '&		"'& word "  HNCword -> dictionary   !dp	THREDING */
  
asm	("
	1:
	.byte 2
	.byte 1
	.byte 0
	.byte 0
	.ascii \"'&\"
	.align 4, 0
	.int  0b
	.equ	ticappendCFA, .
	");

			/* put HNC of "word" in thread		*/
asm	(" .int subCFA	");

  asm	(" .int tickCFA ");

asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");

asm	(" .int trueCFA ");
		/* tell threader "keep going"		*/
asm	(" .int ReturnCFA ");



 /* tttttttttttttttttttttttt============ */

/* '&code		"'&code  word " CFAword->dictionary !dp	 THREDING */
  
asm	("
	0:
	.byte 6
	.byte 1
	.byte 0
	.byte 0
	.ascii \"'&code\"
	.align 4, 0
	.int  1b
	.equ	ticappendcodeCFA, .
	");

			/* put CFA  of "word" in thread		*/
asm	(" .int subCFA	");

  asm	(" .int tickCFA ");

asm	(" .int TOcodeCFA ");

asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");

asm	(" .int trueCFA ");
		/* tell threader "keep going"		*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* 'p&		"'p& word " thread does ->  (P --- HNCword ) !dp THREDING */
  
asm	("
	1:
	.byte 3
	.byte 1
	.byte 0
	.byte 0
	.ascii \"'p&\"
	.align 4, 0
	.int  0b
	.equ	ticpappendCFA, .
	");


asm	(" .int pliteralCFA ");
		/* put pliteral xt in thread		*/
  asm	(" .int pliteralCFA ");
		/* pliteral's CFA as arg to pliteral	*/
asm	(" .int subCFA	");

  asm	(" .int appendpCFA ");

asm	(" .int pdropCFA ");


asm	(" .int subCFA	");
			/* put HNC of "word" in thread		*/
  asm	(" .int ticappendCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============				*/

/* ->C		(P --- tb ) "->C string " 0-term. str. in situ */
  
asm	("
	0:
	.byte 3
	.byte 0
	.byte 0
	.byte 0
	.ascii \"->C\"
	.align 4, 0
	.int  1b
	.equ	TO_CCFA, .
	");

asm	(" .int subCFA	");

  asm	(" .int tokenCFA ");
		/* (got_one) 
asm	(" .int dropCFA ");
		/* () 					*/


SUSPECT

asm	(" .int subCFA	");

  asm	(" .int zeroterminateCFA ");
	/* ()					*/ 
asm	(" .int subCFA	");

  asm	(" .int tbCFA ");
		/* (P tb)				*/
asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============			*/

/* words		( cnt --- ) print cnt wordnames from latest */
  
asm	("
	1:
	.byte 5
	.byte 0
	.byte 0
	.byte 0
	.ascii \"words\"
	.align 4, 0
	.int  0b
	.equ	wordsCFA, .
	");

		/* put downcounter on rs and enter dictionary
		have to size up to cells for un-used bit check	*/
asm	(" .int sTOrCFA ");

asm	(" .int cells_bdCFA ");

asm	(" .int subCFA	");

   asm	(" .int latestCFA ");
			/* (P &latest ) (count)		*/
asm	(" .int pfetchCFA ");
			/* (P latest )	 (count)	*/
asm	(" .int TOrCFA ");
				/* (P latest )	 (R;  count )	*/
asm	("	.equ nwords_one, .	");

	asm	(" .int fetchCFA ");
		/* (P &HNC')	(R;  count ) (*HNC)	*/
		asm	(" .int noCFA ");
			/* (P &HNC')   (R;  count ) ()	*/
		asm	(" .int	 .  + 17 *4  ");
	/* aardvark	*/
	asm	(" .int fetchCFA ");
		/* (P &HNC')   (R;  count ) (*HNC)	*/
	asm	(" .int literalCFA ");

	 asm	(" .int 4	");

	 asm	(" .int 0x01000000	");

	asm	(" .int ANDCFA ");
			/* (P &HNC')   (R;  count ) (userflag)	*/
	asm	(" .int flagCFA ");
		/* (P &HNC')   (R;  count ) (userflag)	*/
	asm	(" .int aintCFA ");

		asm	(" .int noCFA ");
		/* (P &HNC')   (R;  count ) ()		*/
		asm	(" .int	 .  + 5 *4  ");
	/*  NON-USER thread	*/
				/* regular word */
	asm	(" .int subCFA	");

	   asm	(" .int printnameCFA ");

	asm	(" .int subCFA	");

	  asm	(" .int blankCFA ");
	
		asm	(" .int TOlinkCFA ");

		asm	(" .int pfetchCFA ");
	/* (P &HNC')	(R;  count )		*/
		asm	(" .int minusaskrCFA ");
	/* (P &HNC')   (R;  count' )	(flag)	*/
asm	(" .int yesCFA ");
				/* (P &HNC')   (R;  count' )()		*/
asm	("	.int nwords_one	");
			/*  not done with count			*/

asm	(" .int rdropCFA ");
			/* count is 0, ciao	*/
asm	(" .int pdropCFA ");

asm	(" .int rTOsCFA ");

asm	(" .int subCFA	");

 asm	(" .int crCFA ");

asm	(" .int ReturnCFA ");



 /* ttttttttttttttttttttttt============ */

/* 0-terminate	() 	C-ify a token in tb.  !tb	*/
  
asm	("
	0:
	.byte 11
	.byte 0
	.byte 0
	.byte 0
	.ascii \"0-terminate\"
	.align 4, 0
	.int  1b
	.equ	zeroterminateCFA, .
	");

asm	(" .int sTOrCFA ");

asm	(" .int bytesCFA ");

asm	(" .int subCFA	");

 asm	(" .int tbCFA ");
				/* (P tb )			*/
asm	(" .int subCFA	");

  asm	(" .int toklenCFA ");
			/* (P tb tl)			*/
asm	(" .int fetchCFA ");
			/* (P tb tl)	(toklen)	*/
asm	(" .int pdropCFA ");
			/* (P tb )	(toklen)	*/
asm	(" .int pluspCFA ");
			/* (P tb+tl)	()		*/		
asm	(" .int zeroCFA ");
			/* (P tb+tl)	( 0 ) 		*/
asm	(" .int storeCFA ");
			/* (P tb+tl)<--/		*/
asm	(" .int pdropCFA ");
			/* (P )				*/
asm	(" .int rTOsCFA ");

asm	(" .int ReturnCFA ");



 /* ||||||| zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz||============ */

/* zygote		top endless loop, do argv[], then loop over input */
initial_LATEST:		/* label at HNC for dist. init. */
  
asm	("
	1:
	.byte 6
	.byte 0
	.byte 0
	.byte 1
	.ascii \"zygote\"
	.align 4, 0
	.int  0b
	.equ	zygoteCFA, .
	");

		/* Head execution address, CFA, H3m square zero.
		The C init code set up the "commandline args" as
		the interpreter input.	 */
zygote:

asm	(" .int onbitsCFA ");


asm	("	.equ top_loop_of_H3sm_one, .	");

	asm	(" .int subCFA	");

	  asm	(" .int tokenCFA ");
		/* (EOI)			*/
		asm	(" .int yesCFA ");
		/* ()				*/
		asm	(" .int	 .  + 5 *4  ");

	asm	(" .int subCFA	");

	  asm	(" .int interpretCFA ");
		/*   user context  ????????	*/
asm	(" .int ellCFA ");

asm	("	.int top_loop_of_H3sm_one	");

		asm	(" .int subCFA	");
		/* file redirections on stdin must be quitted */
		  asm	(" .int sourceCFA ");
	/* could maybe switch to stdaux, 					or close/open 0/tty */
		asm	(" .int zeroCFA ");

		asm	(" .int storeCFA ");

		asm	(" .int pdropCFA ");

		asm	(" .int subCFA	");

		  asm	(" .int crCFA ");

		asm	(" .int subCFA	");

		  asm	(" .int refillCFA ");

		asm	(" .int dropCFA ");

asm	(" .int ellCFA ");

asm	("	.int top_loop_of_H3sm_one	");



 /* ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc	*/

initial_HERE:

NEXT			/* NEXT ourself to the interpreter... */

} /* ...and C is gone. That's my final comment ---> */
